from loguru import logger
from sweepai.core.github_utils import ClonedRepo, temporary_cloned_repo_copy
from sweepai.dataclasses.file_modification_state import FileModificationState
from sweepai.modify.validate.code_validators import format_file
from sweepai.utils.diff import generate_diff


def format_changes(
    modify_files_dict: FileModificationState,
    cloned_repo: ClonedRepo,
):
    with temporary_cloned_repo_copy(cloned_repo) as temp_repo:
        for file_path, file_contents in modify_files_dict.files.items():
            formatted_original_contents = format_file(file_path, file_contents.original_contents, temp_repo.repo_dir)
            lines_diff_from_original = generate_diff(formatted_original_contents, file_contents.original_contents).count("\n")

            min_diff_size_to_format = 0.3 * len(formatted_original_contents.split("\n"))
            if lines_diff_from_original <= max(25, min_diff_size_to_format): # less than 25 lines or 30% of the file
                file_contents.new_contents = format_file(file_path, file_contents.new_contents, temp_repo.repo_dir)
            else:
                logger.warning(f"File {file_path} was not formatted, {lines_diff_from_original} lines different from original")
    return modify_files_dict




import re

from loguru import logger

from sweepai.core.entities import Message
from sweepai.core.llm.chat import Thread

# TODO: add docs and tests later

system_message = """You are a thorough and meticulous AI assistant helping a user search for relevant files in a codebase to resolve a GitHub issue. The user will provide a description of the issue, including any relevant details, logs, or observations. Your task is to:

1. Summary

Summarize the key points of the issue concisely, but also list out any unfamiliar terms, acronyms, or entities mentioned that may require additional context to fully understand the problem space and identify all relevant code areas.

2. Solution

Describe thoroughly in extreme detail what the ideal code fix would look like:
- Dive deep into the low-level implementation details of how you would change each file. Explain the logic, algorithms, data structures, etc. 
- Explicitly call out any helper functions, utility modules, libraries or APIs you would leverage.
- Carefully consider ALL parts of the codebase that could be relevant, including (in decreasing relevance):
  - Database schemas, models
  - Type definitions, interfaces, enums, constants
  - Shared utility code for common operations like date formatting, string manipulation, etc.
  - Database mutators and query logic 
  - User-facing messages, error messages, localization, i18n
  - Exception handling, error recovery, retries, fallbacks
  - API routes, request/response handling, serialization
  - UI components, client-side logic, event handlers
  - Backend services, data processing, business logic
  - Logging, monitoring, metrics, error tracking, observability, o11y
  - Auth flows, session management, encryption
  - Infrastructure, CI/CD, deployments, config
- List out any unfamiliar domain terms to search for to better understand schemas, types, relationships between entities, etc. Finding data models is key.
- Rate limiting, caching and other cross-cutting concerns could be very relevant for issues with scale or performance.

3. Queries

Generate a list of 10 DIVERSE, highly specific, focused "where" queries to use as vector database search queries to find the most relevant code sections to directly resolve the GitHub issue.
- Reference very specific functions, variables, classes, endpoints, etc. using exact names.
- Describe the purpose and behavior of the code in detail to differentiate it. 
- Ask about granular logic within individual functions/methods.
- Mention adjacent code like schemas, configs, and helpers to establish context.
- Use verbose natural language that mirrors the terminology in the codebase.
- Aim for high specificity to pinpoint the most relevant code in a large codebase.

Format your response like this:

<summary>
[Brief 1-2 sentence summary of the key points of the issue]
</summary>

<solution>
[detailed sentences describing what an ideal fix would change in the code and how

Exhaustive list of relevant parts of the codebase that could be used in the solution include:
- [Module, service, function or endpoint 1] 
- [Module, service, function or endpoint 2]
- [etc.]
</solution>

<queries>
<query>Where is the [EXTREMELY specific description of code section 1]?</query>
<query>Where is the [EXTREMELY specific description of code section 2]?</query>
<query>Where is the [EXTREMELY specific description of code section 3]?</query>
...
</queries>

Examples of good queries:
- Where is the function that compares the user-provided password hash against the stored hash from the database in the user-authentication service?
- Where is the code that constructs the GraphQL mutation for updating a user's profile information, and what specific fields are being updated?
- Where are the React components that render the product carousel on the homepage, and what library is being used for the carousel functionality?
- Where is the endpoint handler for processing incoming webhook events from Stripe in the backend API, and how are the events being validated and parsed?
- Where is the function that generates the XML sitemap for SEO, and what are the specific criteria used for determining which pages are included?
- Where are the push notification configurations and registration logic implemented using the Firebase Cloud Messaging library in the mobile app codebase?
- Where are the Elasticsearch queries that power the autocomplete suggestions for the site's search bar, and what specific fields are being searched and returned?
- Where is the logic for automatically provisioning and scaling EC2 instances based on CPU and memory usage metrics from CloudWatch in the DevOps scripts?"""


def generate_multi_queries(input_query: str):
    thread = Thread(
        messages=[
            Message(
                content=system_message,
                role="system",
            )
        ],
    )
    stripped_input = input_query.strip("\n")
    response = thread.chat(
        content=f"<github_issue>\n{stripped_input}\n</github_issue>",
        model="gpt-4o",
        temperature=0.7,  # I bumped this and it improved the benchmarks
    )
    pattern = re.compile(r"<query>(?P<query>.*?)</query>", re.DOTALL)
    queries = []
    for q in pattern.finditer(response):
        query = q.group("query").strip()
        if query:
            queries.append(query)
    logger.debug(f"Generated {len(queries)} queries from the input query.")
    return queries


if __name__ == "__main__":
    input_query = "I am trying to set up payment processing in my app using Stripe, but I keep getting a 400 error when I try to create a payment intent. I have checked the API key and the request body, but I can't figure out what's wrong. Here is the error message I'm getting: 'Invalid request: request parameters are invalid'. I have attached the relevant code snippets below. Can you help me find the part of the code that is causing this error?"
    generate_multi_queries(input_query)



# IMPORTANT: to use the function decorator your function must have the username as the first param
import asyncio
import traceback
from functools import wraps
from typing import Any, Callable

from fastapi import HTTPException, status
from loguru import logger

from sweepai.o11y.event_logger import posthog
from sweepai.o11y.log_utils import log_to_file
from sweepai.utils.str_utils import get_hash


# TODO: clean this up
def endpoint_wrapper(func: Callable[..., Any]):
    """
    Own custom fastapi wrapper to catch exceptions and log them to posthog and raise errors
    """
    if asyncio.iscoroutinefunction(func):

        @wraps(func)
        async def wrapper(*args, **kwargs):
            func_name = func.__name__
            tracking_id = kwargs.get("tracking_id", get_hash())
            username = kwargs.get("username", "")
            repo_name = kwargs.get("repo_name", "")
            file_path = f"chat-{func_name}-{username}-{repo_name}-{tracking_id}"
            with log_to_file(file_path):
                try:
                    result = await func(*args, **kwargs)
                    return result
                except HTTPException as http_exc:
                    # Re-raise HTTPExceptions without modification
                    raise http_exc
                except Exception as e:
                    # Assuming posthog is properly imported and configured
                    posthog.capture(
                        f"{func_name}_endpoint",
                        f"{func_name}_endpoint error",
                        properties={
                            "error": str(e),
                            "traceback": traceback.format_exc(),
                        },
                    )
                    logger.exception(e)
                    raise HTTPException(
                        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                        detail=f"Error: {str(e)}",
                    ) from e

        return wrapper
    else:

        @wraps(func)
        def wrapper(*args, **kwargs):
            func_name = func.__name__
            tracking_id = kwargs.get("tracking_id", get_hash())
            username = kwargs.get("username", "")
            repo_name = kwargs.get("repo_name", "").replace("/", "--")
            file_path = f"chat-{func_name}-{username}-{repo_name}-{tracking_id}"
            with log_to_file(file_path):
                try:
                    result = func(*args, **kwargs)
                    return result
                except HTTPException as http_exc:
                    # Re-raise HTTPExceptions without modification
                    raise http_exc
                except Exception as e:
                    # Assuming posthog is properly imported and configured
                    posthog.capture(
                        f"{func_name}_endpoint",
                        f"{func_name}_endpoint error",
                        properties={
                            "error": str(e),
                            "traceback": traceback.format_exc(),
                        },
                    )
                    logger.exception(e)
                    raise HTTPException(
                        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                        detail=f"Error: {str(e)}",
                    ) from e

        return wrapper



import copy

from loguru import logger

from sweepai.core.entities import FileChangeRequest, Message
from sweepai.core.github_utils import ClonedRepo
from sweepai.core.llm.chat import Thread, continuous_llm_calls
from sweepai.core.llm.convert_openai_anthropic import AnthropicFunctionCall
from sweepai.dataclasses.code_suggestions import StatefulCodeSuggestion
from sweepai.dataclasses.file_modification_state import FileModificationState
from sweepai.dataclasses.llm_state import ModifyLLMState
from sweepai.modify.modify_prompts import LINTER_WARNING_PROMPT
from sweepai.modify.modify_utils import (
    MODEL,
    NO_TOOL_CALL_PROMPT,
    SLOW_MODEL,
    SUBMIT_TASK_MOCK_FUNCTION_CALL,
    changes_made,
    compile_fcr,
    create_user_message,
    get_current_task_index,
    get_replaces_per_fcr,
    handle_function_call,
    instructions,
    modify_tools,
    render_current_task,
    render_plan,
    tasks_completed,
    validate_and_parse_function_call,
)
from sweepai.modify.validate.code_validators import format_file
from sweepai.utils.diff import generate_diff
from sweepai.utils.streamable_functions import Streamable, streamable


def generate_code_suggestions(
    modify_files_dict: FileModificationState,
) -> list[StatefulCodeSuggestion]:
    applied_code_suggestions = []
    for file_path in sorted(modify_files_dict.files):
        if modify_files_dict.file_has_changes(file_path):
            applied_code_suggestions.append(
                StatefulCodeSuggestion(
                    file_path=file_path,
                    original_code=modify_files_dict.get_original_contents(file_path),
                    new_code=modify_files_dict.get_new_contents(file_path),
                    file_contents=modify_files_dict.get_original_contents(file_path),
                    state="done",
                )
            )
    return applied_code_suggestions


modify_stop_sequences = [
    "</make_change>",
    "</create_file>",
    "</submit_task>",
    "</function_call>",
]


@streamable
def modify(
    fcrs: list[FileChangeRequest],
    request: str,
    cloned_repo: ClonedRepo,
    relevant_filepaths: list[str],
    previous_modify_files_dict: FileModificationState = FileModificationState(),
    raise_on_max_iterations: bool = False,
) -> Streamable[dict[str, dict[str, str]]]:
    # join fcr in case of duplicates
    # handles renames in cloned_repo
    # TODO: handle deletions here - it can cause crashes
    if not fcrs:
        return previous_modify_files_dict
    user_message = create_user_message(
        fcrs=fcrs,
        request=request,
        cloned_repo=cloned_repo,
        relevant_filepaths=relevant_filepaths,
        modify_files_dict=previous_modify_files_dict,
    )
    thread = Thread()
    llm_state: ModifyLLMState = ModifyLLMState(
        initial_check_results={},
        done_counter=0,
        request=request,
        plan=render_plan(fcrs),
        current_task=render_current_task(fcrs),
        user_message_index=1,
        user_message_index_chat_logger=1,
        fcrs=fcrs,
        previous_attempt="",
        changes_per_fcr=[get_replaces_per_fcr(fcr) for fcr in fcrs],
        completed_changes_per_fcr=[0 for _ in fcrs],
        attempt_lazy_change=True,
        attempt_count=0,
        visited_set=set(),
        status_messages=[],
        symbol_to_code_map={},
    )
    full_instructions = instructions + modify_tools
    thread.messages = [Message(role="system", content=full_instructions)]
    try:
        if compiled_fcr := compile_fcr(fcr=fcrs[0], index=0, cloned_repo=cloned_repo):
            thread.messages.append(
                Message(
                    role="user",
                    content=f"Here is the initial user request, plan, and state of the code files:\n{user_message}",
                )
            )
            function_calls_string = compiled_fcr
            thread.messages.append(
                Message(role="assistant", content=function_calls_string)  # this will happen no matter what
            )
            # update messages to make it seem as if it called the fcr
            # update state if it's bad
            # TODO: handling logic to be moved out
            function_call: AnthropicFunctionCall = validate_and_parse_function_call(
                function_calls_string, thread
            )  # this will raise if it's bad but compile_fcr should guarantee it's good
        else:
            model = MODEL
            logger.info(f"Using model: {model}")
            file_names = ", ".join([fcr.filename for fcr in fcrs])
            logger.warning(
                f"[YELLOW] Unexpected LLM call: Calling LLM from modifying with model {model} for files: {file_names}"
            )
            function_calls_string = continuous_llm_calls(
                "initial modify",
                thread,
                content=f"Here is the initial user request, plan, and state of the code files:\n{user_message}",
                model=model,
                stop_sequences=modify_stop_sequences,
            )
    except Exception as e:
        logger.error(f"Error in chat_anthropic: {e}")
        return {}
    modify_files_dict = copy.deepcopy(previous_modify_files_dict)
    # this message list is for the chat logger to have a detailed insight into why failures occur
    detailed_chat_logger_messages = [{"role": message.role, "content": message.content} for message in thread.messages]
    # used to determine if changes were made
    previous_modify_files_dict = copy.deepcopy(modify_files_dict)

    for i in range(max(len(fcrs), 5)):
        yield generate_code_suggestions(modify_files_dict)
        function_call = validate_and_parse_function_call(function_calls_string, thread)
        if function_call:
            num_of_tasks_done = tasks_completed(fcrs)
            # note that detailed_chat_logger_messages is meant to be modified in place by handle_function_call
            function_output, modify_files_dict, llm_state = handle_function_call(
                cloned_repo,
                function_call,
                modify_files_dict,
                llm_state,
                chat_logger_messages=detailed_chat_logger_messages,
            )
            logger.info(function_output)
            fcrs = llm_state.fcrs
            if function_output == "DONE":
                # add the diff of all changes to chat_logger
                break
            # bp()
            detailed_chat_logger_messages.append({"role": "user", "content": function_output})

            if modify_files_dict:  # update the state of the LLM
                user_message = create_user_message(
                    fcrs=fcrs,
                    request=request,
                    cloned_repo=cloned_repo,
                    relevant_filepaths=relevant_filepaths,
                    modify_files_dict=modify_files_dict,
                )
                user_message = f"Here is the UPDATED user request, plan, and state of the code changes. REVIEW THIS CAREFULLY!\n{user_message}"
                # state cleanup should only occur after a task has been finished and if a change was made and if a change was made
                current_num_of_tasks_done = tasks_completed(fcrs)
                if (
                    changes_made(modify_files_dict, previous_modify_files_dict)
                    and current_num_of_tasks_done > num_of_tasks_done
                ):
                    # remove the previous user message and add it to the end, do not remove if it is the inital user message
                    thread.messages = thread.messages[:1]
                    detailed_chat_logger_messages = detailed_chat_logger_messages[:1]
                    thread.messages.append(Message(role="user", content=user_message))
                    detailed_chat_logger_messages.append({"role": "user", "content": user_message})
                    # update the index
                    llm_state.user_message_index = len(thread.messages) - 1
                    llm_state.user_message_index_chat_logger = len(detailed_chat_logger_messages) - 1
                previous_modify_files_dict = copy.deepcopy(modify_files_dict)
        else:
            function_output = NO_TOOL_CALL_PROMPT
        try:
            function_calls_string = ""
            compiled_fcr = ""
            current_fcr_index = get_current_task_index(fcrs)
            change_in_fcr_index = llm_state.completed_changes_per_fcr[current_fcr_index]
            max_changes = llm_state.changes_per_fcr[current_fcr_index]
            if change_in_fcr_index >= max_changes:
                function_calls_string = SUBMIT_TASK_MOCK_FUNCTION_CALL.format(
                    justification=f"Task {current_fcr_index} is now complete."
                )
            else:
                # on first attempt of a new task we use the first fcr
                if llm_state.attempt_lazy_change:
                    if compiled_fcr := compile_fcr(fcrs[current_fcr_index], change_in_fcr_index, cloned_repo):
                        function_calls_string = compiled_fcr
                        function_call = validate_and_parse_function_call(
                            function_calls_string, thread
                        )  # this will raise if it's bad but compile_fcr should guarantee it's good
                        logger.info(f"Function call:\n{function_call}")
                        # update messages to make it seem as if it called the fcr
                        thread.messages.append(Message(role="assistant", content=function_calls_string))
                # if previous things go wrong we make llm call
                if not function_calls_string:
                    if LINTER_WARNING_PROMPT in function_output:
                        llm_state.attempt_count = 3  # skip to opus if there is a linter warning
                    model = MODEL if llm_state.attempt_count < 3 else SLOW_MODEL
                    file_names = ", ".join([fcr.filename for fcr in fcrs])
                    logger.warning(
                        f"Unexpected LLM call: Calling LLM from modifying with model {model} for files: {file_names}"
                    )
                    function_calls_string = continuous_llm_calls(
                        "attempt fix modify",
                        thread,
                        content=function_output,
                        model=model,
                        stop_sequences=modify_stop_sequences,
                    )
                    if function_calls_string in llm_state.visited_set:
                        if llm_state.attempt_count < 3:
                            logger.warning(
                                f"Function call {function_calls_string} has already been visited, retrying with a different model."
                            )
                            llm_state.attempt_count = 3
                            function_calls_string = continuous_llm_calls(
                                "fallback modify",
                                thread,
                                content=function_output,
                                model=model,
                                stop_sequences=modify_stop_sequences,
                            )
                            if function_calls_string in llm_state.visited_set:
                                logger.warning(
                                    f"Function call {function_calls_string} has already been visited, skipping task {current_fcr_index}."
                                )
                                function_calls_string = SUBMIT_TASK_MOCK_FUNCTION_CALL.format(
                                    justification=f"Skipping task {current_fcr_index} due to too many retries."
                                )
                            else:
                                llm_state.visited_set = set()
                        else:
                            logger.warning(
                                f"Function call {function_calls_string} has already been visited, skipping task {current_fcr_index}."
                            )
                            function_calls_string = SUBMIT_TASK_MOCK_FUNCTION_CALL.format(
                                justification=f"Skipping task {current_fcr_index} due to too many retries."
                            )
        except Exception as e:
            logger.error(f"Error in chat_anthropic: {e}")
            with open("msg.txt", "w") as f:
                for message in thread.messages:
                    f.write(f"{message.content}\n\n")
            break
    else:
        logger.error("Max iterations reached")
        if raise_on_max_iterations:
            raise Exception("Max iterations reached")
    for file_path in modify_files_dict.files:
        # this is the diff count of the original code and the formatted code (will be big if the customer doesn't typically using a formatter)
        original_contents = modify_files_dict.get_original_contents(file_path)
        formatted_original_contents = format_file(file_path, original_contents, cloned_repo.repo_dir)
        # if the diff is small, then the original code was already formatted
        original_code_formatted = len(generate_diff(original_contents, formatted_original_contents)) < 10
        if original_code_formatted:  # if the original code was formatted
            formatted_new_contents = format_file(
                file_path,
                modify_files_dict.get_new_contents(file_path),
                cloned_repo.repo_dir,
            )
            modify_files_dict.update_new_contents(file_path, formatted_new_contents)

    diff_string = ""
    for file_name in modify_files_dict.files:
        if diff := generate_diff(
            modify_files_dict.get_original_contents(file_name),
            modify_files_dict.get_new_contents(file_name),
        ):
            diff_string += f"\nChanges made to {file_name}:\n{diff}"
    logger.info(
        "\n".join(
            generate_diff(
                modify_files_dict.get_original_contents(file_name),
                modify_files_dict.get_new_contents(file_name),
            )
            for file_name in modify_files_dict.files
        )
    )  # adding this as a useful way to render the diffs
    yield generate_code_suggestions(modify_files_dict)
    return modify_files_dict


if __name__ == "__main__":
    pass



import re

from sweepai.core.entities import FileChangeRequest, Snippet

# from sweepai.utils.previous_diff_utils import get_relevant_commits

BOT_ANALYSIS_SUMMARY = "bot_analysis_summary"
SNIPPET_TOKEN_BUDGET = int(150_000 * 3.5)  # 140k tokens
MAX_SNIPPETS = 15
RELEVANCE_THRESHOLD = 0.125

sandbox_error_prompt = """The following error logs were returned from `{command}`. Make changes to the current file so that it passes this CI/CD command.

```
{error_logs}
```

Edit old_code to pass the CI/CD."""

sandbox_error_prompt_test = """The following error logs were returned from `{command}`. Make changes to the current file so that it passes this CI/CD command.

```
{error_logs}
```

Edit old_code to pass the CI/CD.
1. Analyze the business logic and tests. Identify whether the failure is in the unit tests or business logic.
2a. If the business logic is correct fix the test to return the expected output.
2b. If the business logic has a bug or you are unsure, skip the failing tests with an explanation."""

GHA_PROMPT = """You're working on resolving a GitHub issue but the code changes fail the GitHub Actions.

You are trying to resolve the following GitHub issue:
<original_github_issue>
{problem_statement}
</original_github_issue>

You made some changes, but GitHub Actions failed with the following logs:
<github_actions_logs>
{github_actions_logs}
</github_actions_logs>

You have previously made the following changes. The diffs represent the current state of the file/project:
<changes_made>
{changes_made}
</changes_made>

Fix the above GitHub Actions."""

GHA_PROMPT_WITH_HISTORY = """You're working on resolving a GitHub issue but the code changes fail the GitHub Actions.

You are trying to resolve the following GitHub issue:
<original_github_issue>
{problem_statement}
</original_github_issue>

Previously the Githu Actions were failing with these logs:
<previous_github_actions_logs>
{previous_github_actions_logs}
</previous_github_actions_logs>

You made some changes to address the previous Github Action failures, but GitHub Actions are now failing with the following logs:
<current_github_actions_logs>
{current_github_actions_logs}
</current_github_actions_logs>

You have previously made the following changes. The diffs represent the current state of the file/project:
<changes_made>
{changes_made}
</changes_made>

Fix the above GitHub Actions."""

issue_sub_request_system_prompt = """You are a tech lead helping to break down a GitHub issue for an intern to solve. Identify every single one of the user's requests. Be complete. The changes should be atomic.

Guidelines:
- For well-specified issues, where all required steps are already listed, simply break down the issue.
- For less well-specified issues, where the user's requests are vague or incomplete, infer the user's intent and break down the issue accordingly. This means you will need to analyze the existing files and list out all the changes that the user is asking for.
- A sub request should correspond to a code or test change.
- A sub request should not be speculative, such as "catch any other errors", "abide by best practices" or "update any other code". Instead explicitly state the changes you would like to see.
- Tests and error handling will be run automatically in the CI/CD pipeline, so do not mention them in the sub requests.
- Topologically sort the sub requests, such that each sub request only depends on sub requests that come before it. For example, create helper functions before using them."""


def cleanup_fcrs(fcrs_string: str):
    fcrs_string = re.sub(
        r"<original_code(?: file_path=\".*?\")?(?: index=\"\d+\")?>",
        "<original_code>",
        fcrs_string,
    )
    fcrs_string = re.sub(
        r"<new_code(?: file_path=\".*?\")?(?: index=\"\d+\")?>",
        "<new_code>",
        fcrs_string,
    )
    return fcrs_string


def parse_patch_fcrs(fcr_patch_string: str):
    pattern = re.compile(
        r"""<(?P<change_type>[a-z_]+)\s+file=\"(?P<filename>[a-zA-Z0-9/\\\.\[\]\(\)\_\+\- @\{\}]*?)\"\s+index=\"(?P<index>\d+)\">(?P<instructions>.*?)\s*<\/\1>""",
        re.DOTALL,
    )
    drop_pattern = re.compile(r"<drop>(\d+?)</drop>", re.DOTALL)
    matches = []
    for match in pattern.finditer(fcr_patch_string):
        matches.append(
            (
                int(match.group("index")),
                FileChangeRequest(
                    change_type=match.group("change_type"),
                    filename=match.group("filename"),
                    instructions=match.group("instructions"),
                ),
            )
        )
    drops = [int(drop.group(1).strip()) for drop in drop_pattern.finditer(fcr_patch_string)]
    matches.sort(key=lambda x: x[0])
    return drops, [match for match in matches]


def parse_renames(renames_string: str):
    pattern = re.compile(r"<rename>(.*?)</rename>", re.DOTALL)
    old_name_pattern = re.compile(r"<old_name>(.*?)</old_name>", re.DOTALL)
    new_name_pattern = re.compile(r"<new_name>(.*?)</new_name>", re.DOTALL)
    rename_dict = {}
    for match in pattern.finditer(renames_string):
        rename_match = match.group(1)
        old_name = old_name_pattern.search(rename_match).group(1)
        if not old_name:
            continue
        new_name = new_name_pattern.search(rename_match).group(1)
        if old_name.strip() == new_name.strip():
            continue
        rename_dict[old_name.strip()] = new_name.strip()
    return rename_dict


def partition_snippets_if_test(snippets: list[Snippet], include_tests=False):
    if include_tests:
        return [snippet for snippet in snippets if "test" in snippet.file_path]
    return [snippet for snippet in snippets if "test" not in snippet.file_path]



import os
from hashlib import md5

from diskcache import Cache
from loguru import logger
from tqdm import tqdm

from sweepai.config.client import SweepConfig
from sweepai.config.server import CACHE_DIRECTORY
from sweepai.core.entities import Snippet
from sweepai.core.llm.tiktoken_utils import Tiktoken
from sweepai.modify.validate.code_validators import chunk_code
from sweepai.utils.file_utils import read_file_with_fallback_encodings
from sweepai.utils.timer import Timer

chunk_cache = Cache(f"{CACHE_DIRECTORY}/chunk_cache")  # we instantiate a singleton, diskcache will handle concurrency
file_name_cache = Cache(f"{CACHE_DIRECTORY}/file_name_cache")

tiktoken_client = Tiktoken()


def is_actual_code_file(directory: str, file: str, sweep_config: SweepConfig) -> bool:
    """
    is_actual_code_file returns true if the file is likely to be a code file.
    """
    cache_key = directory + file
    if cache_key in file_name_cache:
        return file_name_cache[cache_key]
    result = _filter_file(directory, file, sweep_config)
    file_name_cache[cache_key] = result
    return result


def _filter_file(directory: str, file: str, sweep_config: SweepConfig) -> bool:
    """
    Check if a file should be filtered based on its size and other criteria.

    Args:
        file (str): The path to the file.
        sweep_config (SweepConfig): The configuration object.

    Returns:
        bool: True if the file should be included, False otherwise.
    """
    for ext in sweep_config.exclude_exts:
        if file.endswith(ext):
            return False
    only_file_name = file[len(directory) + 1 :]
    only_file_name_parts = only_file_name.split(os.path.sep)
    for dir_name in sweep_config.exclude_dirs:
        for file_part in only_file_name_parts[:-1]:
            if file_part == dir_name:
                return False
    for dir_name in sweep_config.exclude_path_dirs:
        if dir_name in only_file_name_parts:
            return False
    try:
        size = os.stat(file).st_size
        if size > 240000 or size < 10:
            return False
    except FileNotFoundError as e:
        logger.info(f"File not found: {file}. {e}")
        return False
    if not os.path.isfile(file):
        return False
    try:
        data = read_file_with_fallback_encodings(file)
    except UnicodeDecodeError:
        logger.warning(f"UnicodeDecodeError: {file}, skipping")
        return False
    if b"\x00" in data.encode():
        return False
    line_count = data.count("\n") + 1
    # if average line length is greater than 200, then it is likely not human readable
    if len(data) / line_count > 200:
        return False
    # check token density, if it is greater than 2, then it is likely not human readable
    token_count = tiktoken_client.count(data[:1000])
    if token_count == 0:
        return False
    if len(data[:1000]) / token_count < 2 and len(data) > 100:
        return False
    return True


def read_file(file_name: str) -> str:
    try:
        with open(file_name, "r") as f:
            return f.read()
    except Exception:
        return ""


FILE_THRESHOLD = 240


def conditional_hash(contents: str):
    if len(contents) > 255:
        return md5(contents.encode()).hexdigest()
    return contents


def file_path_to_chunks(file_path: str) -> list[str]:
    file_contents = read_file(file_path)
    content_hash = conditional_hash(file_path + file_contents)
    if content_hash in chunk_cache:
        return chunk_cache[content_hash]
    chunks = chunk_code(file_contents, path=file_path)
    chunk_cache[content_hash] = chunks
    return chunks


# @file_cache()
def directory_to_chunks(
    repo_dir: str,
    sweep_config: SweepConfig,
    do_not_use_file_cache: bool = False,
) -> tuple[list[Snippet], list[str]]:
    # dir_file_count = {}

    logger.info(f"Reading files from {repo_dir}")
    vis = set()

    # 81.5s -> 42.68
    def dfs(file_path: str = repo_dir):
        only_file_name = os.path.basename(file_path)
        if only_file_name in ("node_modules", ".venv", "venv", "patch"):
            return
        if file_path in vis:
            return
        vis.add(file_path)
        try:
            with os.scandir(file_path) as it:
                children = list(it)
                if len(children) > FILE_THRESHOLD:
                    return
                for entry in children:
                    if entry.is_dir(follow_symlinks=False):
                        yield from dfs(entry.path)
                    else:
                        yield entry.path
        except NotADirectoryError:
            yield file_path

    with Timer():
        file_list = dfs()
        file_list = [
            file_name
            for file_name in tqdm(file_list)
            if is_actual_code_file(repo_dir, file_name, sweep_config)
            # and os.path.isfile(file_name) # should be unneeded
        ]
    logger.info("Done reading files")
    all_chunks = []
    # TODO: It's worth refactoring this to check the cache first sequentially, then potentially multiprocessing if there are enough misses
    # with multiprocessing.Pool(processes=multiprocessing.cpu_count() // 4) as pool:
    #     for chunks in tqdm(pool.imap(file_path_to_chunks, file_list), total=len(file_list), desc="Chunking files"):
    #         all_chunks.extend(chunks)
    # sequential for testing
    for file_name in tqdm(file_list, desc="Chunking files"):
        all_chunks.extend(file_path_to_chunks(file_name))
    return all_chunks, file_list


if __name__ == "__main__":
    try:
        from sweepai.core.github_utils import ClonedRepo, get_installation_id

        organization_name = "sweepai"

        installation_id = get_installation_id(organization_name)
        cloned_repo = ClonedRepo("sweepai/sweep", installation_id, "main")
        sweep_config = SweepConfig()
        chunks, file_list = directory_to_chunks(cloned_repo.repo_dir, sweep_config)
        # ensure no unallowed files are let through
        assert not any([file for file in file_list if sweep_config.is_file_excluded(file)])
        # pick 10 random files and turn them to chunks
        import random

        for _ in range(10):
            idx = random.randint(0, len(file_list) - 1)
            file_chunks = file_path_to_chunks(file_list[idx])

    except Exception as e:
        logger.error(f"repo_parsing_utils.py failed to run successfully with error: {e}")



import io
import re
import ssl
import zipfile
from contextlib import suppress
from functools import lru_cache
from time import sleep
from urllib.parse import urlparse, urlunparse

import requests
from api4jenkins import Jenkins
from api4jenkins.exceptions import AuthenticationError
from github.CheckRun import CheckRun
from github.Commit import Commit
from github.CommitStatus import CommitStatus
from github.PullRequest import PullRequest
from loguru import logger
from pydantic import BaseModel

from sweepai.config.server import (
    CURL_CA_BUNDLE,
    DEV,
    GITHUB_API_BASE_URL,
    JENKINS_AUTH,
)
from sweepai.core.github_utils import get_token
from sweepai.utils.cache import create_cache
from sweepai.utils.ticket_rendering_utils import (
    hyperlink,
)

logs_cache = create_cache()

### UTILS

# https://docs.github.com/en/rest/checks/runs?apiVersion=2022-11-28#create-a-check-run--parameters
CONCLUSION_TYPES = [
    "action_required",
    "cancelled",
    "failure",
    "neutral",
    "success",
    "skipped",
    "stale",
    "timed_out",
]
STATUS_TYPES = ["completed", "in_progress", "queued", "waiting", "requested", "pending"]
MIN_WAIT_SECONDS = 600  # 10 minutes

RunType = CommitStatus | CheckRun


def create_hyperlink(run: RunType | Commit):
    if isinstance(run, CommitStatus):
        return hyperlink(f"{run.context} - {run.description}", run.target_url)
    elif isinstance(run, CheckRun):
        return hyperlink(f"{run.name} - {run.conclusion}", run.details_url)
    elif isinstance(run, Commit):
        return hyperlink(f"{run.sha[:7]}", run.html_url)
    raise Exception("Invalid run type")


class FixCIInput(BaseModel):
    cleaned_logs: str
    commit_hash: str


def interval_timer(
    total_time: int = 60 * 30,
    interval: int = 5,
):
    for i in range(total_time // interval):
        yield i
        sleep(interval)


### JENKINS LOGS


@lru_cache
def get_jenkins_client(base_url: str):
    if not JENKINS_AUTH:
        return None
    kwargs = (
        {
            "verify": ssl.create_default_context(cafile=CURL_CA_BUNDLE),
            "proxies": {"all://": None},
        }
        if CURL_CA_BUNDLE
        else {}
    )
    for auth in JENKINS_AUTH:
        with suppress(AuthenticationError):
            client = Jenkins(base_url, auth=auth, **kwargs)
            logger.info(client.version)
            break
    else:
        logger.warning("No Jenkins auth provided, skipping Jenkins CI failure check")
        return None
    return client


# E.g. foo » bar » repo » build-pipeline #123
TRIGGERED_BUILDS_PATTERN = r"^(?P<job_name>(.*? » )+)(?P<build_name>.*?) #(?P<build_number>\d+):"


def extract_triggered_builds(build_logs: str) -> list[str]:
    links = []
    sep = " » "
    for match_ in re.finditer(TRIGGERED_BUILDS_PATTERN, build_logs, re.MULTILINE):
        path = match_.group("job_name").strip(sep).split(sep)
        path.append(match_.group("build_name"))
        links.append("/job/" + "/job/".join(path) + "/" + match_.group("build_number"))
    return links


def fetch_jenkins_build_logs(target_url: str):
    parsed_url = urlparse(target_url)
    if not (client := get_jenkins_client(f"{parsed_url.scheme}://{parsed_url.netloc}")):
        logger.warning("No Jenkins auth provided, skipping Jenkins CI failure check")
        return ""
    # Jenkins Python client has bugs
    response = requests.get(target_url.replace("display/redirect", "consoleText"), auth=client._auth)
    try:
        response.raise_for_status()
        original_logs = response.text
        # Sometimes the builds are seed builds that trigger other builds. We search one level deep (no recursion).
        triggered_build_paths = extract_triggered_builds(original_logs)
        for build_path in triggered_build_paths:
            url = urlunparse(
                (
                    parsed_url.scheme,
                    parsed_url.netloc,
                    build_path + "/consoleText",
                    "",
                    "",
                    "",
                )
            )
            response = requests.get(url, auth=client._auth)
            if response.ok:
                original_logs += "\n\n" + response.text
            else:
                logger.error(f"Failed to fetch Jenkins build logs at {build_path}: {response.status_code}")
        return original_logs
    except requests.exceptions.HTTPError as e:
        logger.error(f"Failed to fetch Jenkins build logs: {e}")
        return ""


### GHA LOGS


def get_run_id(run: CheckRun):
    # E.g. https://github.com/sweepai/sweep-internal/actions/runs/11171470435/job/31056237554
    # PyGithub's "run.id" is actually the job id, run_id is the workflow id, confusing, I know
    _, _, _, _, run_id, *_ = urlparse(run.details_url).path.strip("/").split("/")
    return run_id


def truncate_logs(logs_str: str):
    log_list = logs_str.split("\n")
    truncated_logs = [log[log.find(" ") + 1 :] for log in log_list]
    logs_str = "\n".join(truncated_logs)
    return logs_str


def clean_ansi_logs(logs_str: str) -> str:  # TODO: unit test
    ansi_escape = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", logs_str)


def format_logs(logs_str: str) -> str:  # TODO: unit test
    return (
        logs_str.replace("##[error]", "\n")
        .replace("##[endgroup]", "\nThe relevant logs are:\n")
        .replace("##[group]", "The command that fails:\n\n")
    )


def post_process_logs(logs_str: str) -> str:  # TODO: unit test
    logs_str = truncate_logs(logs_str)
    logs_str = clean_ansi_logs(logs_str)
    logs_str = format_logs(logs_str)
    return logs_str


def unzip_and_extract_logs(blob: bytes, get_errors_only=True) -> str:
    # Only fetches relevant logs from the zip file
    logs_str = ""
    zip_file = zipfile.ZipFile(io.BytesIO(blob))
    for file in zip_file.namelist():
        if file.endswith(".txt"):
            with zip_file.open(file) as f:
                logs = f.read().decode("utf-8")
                last_line = logs.splitlines()[-1]
                if "##[error]" in last_line:
                    logs_str += logs
                elif not get_errors_only:  # get all logs
                    logs_str += logs
    return logs_str


def download_raw_logs(repo_full_name: str, run_id: str, installation_id: int, get_errors_only=True) -> bytes:
    for _ in range(3):
        # There's actually a hidden redirect here, and it gets automatically handled by the requests library
        headers = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {get_token(installation_id=installation_id)}",
            "X-GitHub-Api-Version": "2022-11-28",
        }
        response = requests.get(
            f"{GITHUB_API_BASE_URL}/repos/{repo_full_name}/actions/runs/{run_id}/logs",
            headers=headers,
        )
        # check for 404
        if response.status_code == 404:
            continue
        return response.content
    raise Exception(f"Failed to download logs for {repo_full_name} run {run_id}")


### ROUTER


def get_ci_failure_logs(installation_id: int, runs: list[RunType], pull_request: PullRequest):
    logs_list = []
    fetched_urls = set()
    for run in runs:
        logger.info(run)
        if isinstance(run, CommitStatus):
            logger.info(f"Run context {run.context}")
            # WARNING: run.target_url might not have jenkins in the url if the organization names it strangely
            # (if so, make this a config) 
            if "jenkins" in run.target_url:  # sometimes it's /pr-head sometimes it's /branch
                # Jenkins format: http://{BASE_URL}/job/{JOB_NAME}/job/PR-{PR_NUMBER}/{RUN_NUMBER}/display/redirect
                if run.target_url not in fetched_urls:
                    logs = fetch_jenkins_build_logs(run.target_url)
                    if logs:
                        logs_list.append(logs)
                        fetched_urls.add(run.target_url)
            # TODO: add support for other CI providers
        elif isinstance(run, CheckRun):
            repo_full_name = pull_request.base.repo.full_name
            actual_run_id = get_run_id(run)
            cache_key = (repo_full_name, actual_run_id, "v3")
            if DEV and (cached_logs := logs_cache.get(cache_key)):
                logs_list.append(cached_logs)
            else:
                raw_binary_logs = download_raw_logs(repo_full_name=repo_full_name, run_id=actual_run_id, installation_id=installation_id)
                raw_logs = unzip_and_extract_logs(raw_binary_logs)
                logs = post_process_logs(raw_logs)
                logs_list.append(logs)
                logs_cache.set(cache_key, logs)
    # TODO: add support for docker runs
    return logs_list


### CI FETCHING LOGIC


def get_ci_failures(pull_request: PullRequest) -> tuple[list[RunType], list[RunType], list[RunType]]:
    commit: Commit = pull_request.get_commits().reversed[0]
    commit_statuses = list(commit.get_statuses())
    pending_statuses = [status for status in commit_statuses if status.state == "pending"]
    successful_statuses = [status for status in commit_statuses if status.state == "success"]
    failing_statuses = [
        status for status in commit_statuses if status.state in ("failure", "error")
    ]  # idfk why github has both failure and error
    check_runs = list(commit.get_check_runs())
    pending_check_runs = [run for run in check_runs if run.status != "completed"]
    successful_check_runs = [run for run in check_runs if run.conclusion == "success"]
    failing_check_runs = [run for run in check_runs if run.conclusion == "failure"]
    return pending_statuses + pending_check_runs, successful_statuses + successful_check_runs, failing_statuses + failing_check_runs


def wait_for_ci_failures(pull_request: PullRequest) -> tuple[list[RunType], list[RunType], list[RunType]]:
    # can use async
    pending_runs = []
    successful_runs = []
    error_runs = []
    for i in interval_timer():
        pending_runs, successful_runs, error_runs = get_ci_failures(pull_request)

        if error_runs:
            break
        elif not pending_runs and i > MIN_WAIT_SECONDS // 5:
            return [], [], []
        else:
            logger.debug(f"Pending runs: {pending_runs}")
            logger.debug(f"Successful runs: {successful_runs}")
            logger.debug(f"PR URL: {pull_request.html_url}")
    return pending_runs, successful_runs, error_runs



# pylint: disable=all
import sys
from bdb import BdbQuit
from contextlib import contextmanager


@contextmanager
def post_mortem():
    try:
        yield
    except BdbQuit:
        raise
    except Exception as e:
        import pdb  # noqa: T100

        pdb.post_mortem()
        raise e


@contextmanager
def bp():
    try:
        yield
    except BdbQuit:
        raise
    except Exception as e:
        # goes 3 frames back
        import pdb  # noqa: T100

        frame = sys._getframe()
        for _ in range(2):
            if frame.f_back:
                frame = frame.f_back
        pdb.Pdb().set_trace(frame)
        raise e



"""
on_ticket is the main function that is called when a new issue is created.
It is only called by the webhook handler in sweepai/api.py.
"""

import hashlib
import io
import os
import re
import traceback
import zipfile
from bdb import BdbQuit
from contextlib import contextmanager
from dataclasses import dataclass
from itertools import chain
from typing import Callable

import requests
from github import IncompletableObject
from github.Issue import Issue
from github.IssueComment import IssueComment
from github.PullRequest import PullRequest
from github.PullRequestComment import PullRequestComment
from loguru import logger

from sweepai.config.client import SweepConfig
from sweepai.config.server import DOCUMENTATION_URL
from sweepai.core.github_utils import get_token
from sweepai.dataclasses.codereview import CodeReview, CodeReviewIssue
from sweepai.handlers.on_check_suite import clean_gh_logs, remove_ansi_tags
from sweepai.utils.str_utils import sanitize_string_for_github

MESSAGE_MARKER = "\n\n<!--- MESSAGE_MARKER --->"
PR_REVERT_BUTTON = "- [ ] Revert"
PR_REVERTED_INDICATOR = "- [x] Revert"
DOCUMENTATION_STRING = (
    f"\n\n:book: For more information on how to use Sweep, please read our [documentation]({DOCUMENTATION_URL})."
)

sweeping_png = """<a href="https://sweep.dev"><img src="https://raw.githubusercontent.com/sweepai/sweep/main/.assets/sweep-square.png" width="100" style="width:50px; margin-bottom:10px" alt="Sweeping"></a>"""

INSTRUCTIONS_FOR_REVIEW = """\
### 💡 To get Sweep to edit this pull request, you can:
* Comment below, and Sweep can edit the entire PR
* Comment on a file, Sweep will only modify the commented file
* Edit the original issue to get Sweep to recreate the PR from scratch"""

SWEEP_PR_REVIEW_HEADER = "# Sweep: PR Review"

HR = "\n\n---\n\n"

OVERRIDE_ISSUE_VALIDATION = "OVERRIDE_ISSUE_VALIDATION"


def center(text: str) -> str:
    return f"<div align='center'>{text}</div>"


SWEEPING_PNG = f"{center(sweeping_png)}\n\n"


def hyperlink(text: str, url: str) -> str:
    return f"[{text}]({url})"


# Add :eyes: emoji to ticket
def add_emoji(issue: Issue, comment_id: int = None, reaction_content="eyes"):
    item_to_react_to = issue.get_comment(comment_id) if comment_id else issue
    item_to_react_to.create_reaction(reaction_content)


# Add :eyes: emoji to ticket
def add_emoji_to_pr(pr: PullRequest, comment_id: int = None, reaction_content="eyes"):
    item_to_react_to = pr.get_comment(comment_id) if comment_id else pr
    item_to_react_to.create_reaction(reaction_content)


# takes in a list of workflow runs and returns a list of messages containing the logs of the failing runs
def get_failing_gha_logs(runs, installation_id) -> str:
    token = get_token(installation_id)
    all_logs = ""
    for run in runs:
        # jobs_url
        jobs_url = run.jobs_url
        jobs_response = requests.get(
            jobs_url,
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {token}",
                "X-GitHub-Api-Version": "2022-11-28",
            },
        )
        if jobs_response.status_code == 200:
            failed_jobs = []
            jobs = jobs_response.json()["jobs"]
            for job in jobs:
                if job["conclusion"] == "failure":
                    failed_jobs.append(job)

            failed_jobs_name_list = []
            for job in failed_jobs:
                # add failed steps
                for step in job["steps"]:
                    if step["conclusion"] == "failure":
                        parsed_name = step["name"].replace("/", "")
                        failed_jobs_name_list.append(f"{job['name']}/{step['number']}_{parsed_name}")
        else:
            logger.error("Failed to get jobs for failing github actions, possible a credentials issue")
            return all_logs
        # make sure jobs in valid
        if jobs_response.json()["total_count"] == 0:
            logger.warning(f"no jobs for this run: {run}, continuing...")
            continue

        # logs url
        logs_url = run.logs_url
        logs_response = requests.get(
            logs_url,
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {token}",
                "X-GitHub-Api-Version": "2022-11-28",
            },
            allow_redirects=True,
        )
        # Check if the request was successful
        if logs_response.status_code == 200:
            zip_data = io.BytesIO(logs_response.content)
            zip_file = zipfile.ZipFile(zip_data, "r")
            zip_file_names = zip_file.namelist()
            for file in failed_jobs_name_list:
                if f"{file}.txt" in zip_file_names:
                    logs = zip_file.read(f"{file}.txt").decode("utf-8")
                    logs_prompt = clean_gh_logs(logs)
                    all_logs += logs_prompt + "\n"
        else:
            logger.error("Failed to get logs for failing github actions, likely a credentials issue")
    return remove_ansi_tags(all_logs)


def get_branch_diff_text(repo, branch, base_branch=None):
    base_branch = base_branch or SweepConfig.get_branch(repo)
    comparison = repo.compare(base_branch, branch)
    file_diffs = comparison.files

    priorities = {
        "added": 0,
        "renamed": 1,
        "modified": 2,
        "removed": 3,
    }

    file_diffs = sorted(file_diffs, key=lambda x: priorities.get(x.status, 4))

    pr_diffs = []
    for file in file_diffs:
        diff = file.patch
        if file.status == "added" or file.status == "modified" or file.status == "removed" or file.status == "renamed":
            pr_diffs.append((file.filename, diff))
        else:
            logger.info(f"File status {file.status} not recognized")  # TODO(sweep): We don't handle renamed files
    return "\n".join([f"{filename}\n{diff}" for filename, diff in pr_diffs])


def parse_issues_from_code_review(issue_string: str):
    issue_regex = r"<issue>(?P<issue>.*?)<\/issue>"
    issue_matches = list(re.finditer(issue_regex, issue_string, re.DOTALL))
    potential_issues = set()
    for issue in issue_matches:
        issue_content = issue.group("issue")
        issue_params = ["issue_description", "file_name", "line_number"]
        issue_args = {}
        issue_failed = False
        for param in issue_params:
            regex = rf"<{param}>(?P<{param}>.*?)<\/{param}>"
            result = re.search(regex, issue_content, re.DOTALL)
            try:
                issue_args[param] = result.group(param).strip()
            except AttributeError:
                issue_failed = True
                break
        if not issue_failed:
            potential_issues.add(CodeReviewIssue(**issue_args))
    return list(potential_issues)


# converts the list of issues inside a code_review into markdown text to display in a github comment
def render_code_review_issues(
    username: str,
    pr: PullRequest,
    code_review: CodeReview,
    issue_type: str = "",
    sorted_issues: list[CodeReviewIssue] = [],  # changes how issues are rendered
):
    files_to_blobs = {file.filename: file.blob_url for file in list(pr.get_files())}
    # generate the diff urls
    files_to_diffs = {}
    for file_name, _ in files_to_blobs.items():
        sha_256 = hashlib.sha256(file_name.encode("utf-8")).hexdigest()
        files_to_diffs[file_name] = f"{pr.html_url}/files#diff-{sha_256}"
    if sorted_issues:
        code_issues = code_review.issues
    else:
        code_issues = code_review.issues
    if issue_type == "potential":
        code_issues = code_review.potential_issues
    code_issues_string = ""
    for issue in code_issues:
        if issue.file_name in files_to_blobs:
            issue_blob_url = f"{files_to_blobs[issue.file_name]}#L{issue.line_number}"
            issue_diff_url = f"{files_to_diffs[issue.file_name]}R{issue.line_number}"
            if sorted_issues:
                code_issues_string += f"<li>In `{issue.file_name}`: {issue.issue_description}</li>\n\n{issue_blob_url}\n[View Diff]({issue_diff_url})"
            else:
                code_issues_string += (
                    f"<li>{issue.issue_description}</li>\n\n{issue_blob_url}\n[View Diff]({issue_diff_url})"
                )
    return code_issues_string


def escape_html(text: str) -> str:
    return text.replace("<", "&lt;").replace(">", "&gt;")


# make sure code blocks are render properly in github comments markdown
def format_code_sections(text: str) -> str:
    backtick_count = text.count("`")
    if backtick_count % 2 != 0:
        # If there's an odd number of backticks, return the original text
        return text
    result = []
    last_index = 0
    inside_code = False
    while True:
        try:
            index = text.index("`", last_index)
            result.append(text[last_index:index])
            if inside_code:
                result.append("</code>")
            else:
                result.append("<code>")
            inside_code = not inside_code
            last_index = index + 1
        except ValueError:
            # No more backticks found
            break
    result.append(text[last_index:])
    formatted_text = "".join(result)
    # Escape HTML characters within <code> tags
    formatted_text = formatted_text.replace("<code>", "<code>").replace("</code>", "</code>")
    parts = formatted_text.split("<code>")
    for i in range(1, len(parts)):
        code_content, rest = parts[i].split("</code>", 1)
        parts[i] = escape_html(code_content) + "</code>" + rest

    return "<code>".join(parts)


def create_review_comments_for_code_issues(pr: PullRequest, code_issues: list[CodeReviewIssue]):
    commit_sha = pr.head.sha
    commits = list(pr.get_commits())
    pr_commit = None
    for commit in commits:
        if commit.sha == commit_sha:
            pr_commit = commit
            break
    for issue in code_issues:
        comment_body = issue.issue_description
        comment_line = int(issue.line_number)
        comment_path = os.path.normpath(issue.file_name)
        pr.create_review_comment(body=comment_body, commit=pr_commit, path=comment_path, line=comment_line)


# turns code_review_by_file into markdown string
def render_pr_review_by_file(
    username: str,
    pr: PullRequest,
    code_review_by_file: dict[str, CodeReview],
    formatted_comment_threads: dict[str, str],
    pull_request_summary: str = "",
    dropped_files: list[str] = [],
    unsuitable_files: list[tuple[str, Exception]] = [],
    pr_authors: str = "",
) -> str:
    body = f"{SWEEP_PR_REVIEW_HEADER}\n"
    pr_summary = ""
    if pr_authors:
        body += f"Authors: {pr_authors}\n" if ", " in pr_authors else f"Author: {pr_authors}\n"
    # pull request summary goes to the bottom
    if pull_request_summary:
        pr_summary += f"\n<h3>Summary</h3>\n{pull_request_summary}\n<hr>\n"
    issues_section = ""
    code_issues_section = ""
    potential_issues_section = ""
    # build issues section
    # create review comments for all the issues
    all_issues = []
    all_potential_issues = []
    for _, code_review in code_review_by_file.items():
        all_issues.extend(code_review.issues)
        all_potential_issues.extend(code_review.potential_issues)
    create_review_comments_for_code_issues(pr, all_issues)
    # build issues section
    for file_name, code_review in code_review_by_file.items():
        code_issues = code_review.issues
        if code_issues:
            code_issues_string = render_code_review_issues(username, pr, code_review)
            code_issues_section += f"""<details>
<summary>{file_name}</summary>
<ul>{format_code_sections(code_issues_string)}</ul></details>"""
    # build potential issues section
    for file_name, code_review in code_review_by_file.items():
        potential_issues = code_review.potential_issues
        if potential_issues:
            potential_issues_string = render_code_review_issues(username, pr, code_review, issue_type="potential")
            potential_issues_section += f"""<details>
<summary>{file_name}</summary>
<ul>{format_code_sections(potential_issues_string)}</ul></details>"""
    # add titles/dropdowns for issues and potential issues section depending on if there were any issues/potential issues
    if code_issues_section:
        code_issues_section = (
            f"<details><summary><h3>Issues</h3></summary><p><strong>Sweep found these issues in your code.</strong></p>\n\n{code_issues_section}</details><hr>"
            ""
        )
    if potential_issues_section:
        potential_issues_section = f"<details><summary><h3>Potential Issues</h3></summary><p><strong>Sweep is unsure if these are issues, but they might be worth checking out.</strong></p>\n\n{potential_issues_section}</details><hr>"
    # add footer describing dropped files
    footer = ""
    if len(dropped_files) == 1:
        footer += f"<p>{dropped_files[0]} was not reviewed because our filter identified it as typically a non-human-readable (auto-generated) or less important file (e.g., dist files, package.json, images). If this is an error, please let us know.</p>"
    elif len(dropped_files) > 1:
        dropped_files_string = "".join([f"<li>{file}</li>" for file in dropped_files])
        footer += f"<p>The following files were not reviewed because our filter identified them as typically non-human-readable (auto-generated) or less important files (e.g., dist files, package.json, images). If this is an error, please let us know.</p><ul>{dropped_files_string}</ul>"
    if len(unsuitable_files) == 1:
        footer += f"<p>The following file {unsuitable_files[0][0]} were not reviewed as they were deemed unsuitable for the following reason: {str(unsuitable_files[0][1])}. If this is an error please let us know.</p>"
    elif len(unsuitable_files) > 1:
        unsuitable_files_string = "".join(
            [f"<li>{file}: {str(exception)}</li>" for file, exception in unsuitable_files]
        )
        footer += f"<p>The following files were not reviewed as they were deemed unsuitable for a variety of reasons. If this is an error please let us know.</p><ul>{unsuitable_files_string}</ul>"
    if len(all_issues) == 0 and len(all_potential_issues) == 0:
        issues_section = "The Pull Request looks good! Sweep did not find any issues."
        if not formatted_comment_threads:
            issues_section = "The Pull Request looks good! Sweep did not find any new issues."
    elif len(all_issues) == 0:
        issues_section = "The Pull Request looks good! Sweep did not find any issues but found some potential issues that you may want to take a look at."
        if not formatted_comment_threads:
            issues_section = "The Pull Request looks good! Sweep did not find any new issues but found some potential issues that you may want to take a look at."
    else:
        if len(all_issues) == 1:
            issues_section = f"\n\nSweep found `{len(all_issues)}` new issue.\n\n"
        else:
            issues_section = f"\n\nSweep found `{len(all_issues)}` new issues.\n\n"
        issues_section += "Sweep has left comments on the pull request for you to review. \nYou may respond to any comment Sweep made your feedback will be taken into consideration if you run the review again. If Sweep made a mistake, you can resolve the comment or let Sweep know by responding to the comment."
    return body + issues_section + code_issues_section + potential_issues_section + pr_summary + footer


# handles the creation or update of the Sweep comment letting the user know that Sweep is reviewing a pr
# returns the comment_id
def create_update_review_pr_comment(
    username: str,
    pr: PullRequest,
    formatted_comment_threads: dict[str, str],
    code_review_by_file: dict[str, CodeReview] | None = None,
    pull_request_summary: str = "",
    dropped_files: list[str] = [],
    unsuitable_files: list[tuple[str, Exception]] = [],
    error_message: str = "",  # passing in an error message takes priority over everything else
) -> int:
    comment_id = -1
    sweep_comment = None
    # comments that appear in the github ui in the conversation tab are considered issue comments
    pr_comments = list(pr.get_issue_comments())
    # make sure we don't already have a comment created
    for comment in pr_comments:
        # a comment has already been created
        if comment.body.startswith(SWEEP_PR_REVIEW_HEADER):
            comment_id = comment.id
            sweep_comment = comment
            break
    commits = list(pr.get_commits())
    pr_authors = set()
    try:
        pr_authors.add(f"{pr.user.login}")
    except Exception as e:
        logger.error(f"Failed to retrieve {pr.user}: {str(e)}")
    for commit in commits:
        author = commit.author
        try:
            if author:
                pr_authors.add(f"{author.login}")
        except IncompletableObject as e:
            logger.error(f"Failed to retrieve author {author} for commit {commit.sha}: {str(e)}")
    pr_authors = ", ".join(pr_authors)

    # comment has not yet been created
    if not sweep_comment:
        comment_content = f"{SWEEP_PR_REVIEW_HEADER}\nSweep is currently reviewing your pr..."
        if pr_authors:
            comment_content = f"{SWEEP_PR_REVIEW_HEADER}\nAuthors of pull request: {pr_authors}\n\nSweep is currently reviewing your pr..."
        sweep_comment = pr.create_issue_comment(comment_content)

    # update the comment
    if error_message:
        sweep_comment.edit(
            f"{SWEEP_PR_REVIEW_HEADER}\nSweep was unable to review your pull request due to the following reasons:\n\n{error_message}"
        )
        comment_id = sweep_comment.id
        return comment_id  # early return

    # update body of sweep_comment
    if code_review_by_file:
        rendered_pr_review = render_pr_review_by_file(
            username,
            pr,
            code_review_by_file,
            formatted_comment_threads,
            pull_request_summary=pull_request_summary,
            dropped_files=dropped_files,
            unsuitable_files=unsuitable_files,
            pr_authors=pr_authors,
        )
        sweep_comment.edit(rendered_pr_review)
    comment_id = sweep_comment.id
    return comment_id


def format_tracking_id(tracking_id: str | None) -> str:
    if tracking_id:
        return f'\n\n<div align="right"><sup><sup><em>Tracking ID: {tracking_id}</em></sup></sup></div>'
    return ""


def generate_validation_skip_warning() -> str:
    return "> [!WARNING]\n> The issue validation was skipped due to the presence of the `OVERRIDE_ISSUE_VALIDATION` flag. Please ensure that your issue description is clear and actionable."


def generate_revert_pr_comment(commit_message: str, commit_sha: str, pull_request_url: str) -> str:
    formatted_commit_url = f"[{commit_message}]({pull_request_url}/commits/{commit_sha})"
    return (
        f"\n\n# 🔄 Revert {commit_sha}?\nTo revert `{commit_message}` click the checkbox below.\n"
        + f"{PR_REVERT_BUTTON} {formatted_commit_url}"
    )


def get_suffix_for_sweep_comment(tracking_id: str | None, message_marker: str = MESSAGE_MARKER) -> str:
    return DOCUMENTATION_STRING + message_marker + format_tracking_id(tracking_id)

def create_message_marker(marker_id: str) -> str:
    return f"\n\n<!--- MESSAGE_MARKER {marker_id} -->"

def find_sweeps_github_comment(
    issue_or_pr: Issue | PullRequest, message_marker: str = MESSAGE_MARKER
) -> IssueComment | None:
    # NOTE: for pull requests, get_comments only gets review comments, so we need to get issue comments too if it's a pull request
    all_comments = chain(
        issue_or_pr.get_comments(),
        (issue_or_pr.get_issue_comments() if isinstance(issue_or_pr, PullRequest) else []),
    )
    return next((comment for comment in all_comments if message_marker in comment.body), None)


def create_or_edit_sweeps_github_comment(
    issue_or_pr: Issue | PullRequest,
    sweeps_message: str,
    sweeps_github_comment: IssueComment | None = None,
    tracking_id: str | None = None,
    message_marker: str = MESSAGE_MARKER,
) -> IssueComment:
    """
    If this is not used to exit, return the sweeps_github_comment so that the caller can edit it.
    """
    # existing comment exists, edit the comment
    body = sweeps_message + get_suffix_for_sweep_comment(tracking_id, message_marker)
    if sweeps_github_comment:
        sweeps_github_comment.edit(body=body)
    else:
        sweeps_github_comment = find_sweeps_github_comment(issue_or_pr, message_marker)
        if sweeps_github_comment:
            sweeps_github_comment.edit(body=body)
        else:
            # no existing comment exists, create the comment
            if isinstance(issue_or_pr, PullRequest):
                sweeps_github_comment = issue_or_pr.create_issue_comment(body)
            elif isinstance(issue_or_pr, Issue):
                sweeps_github_comment = issue_or_pr.create_comment(body)
    return sweeps_github_comment


@dataclass
class UserFacingError(Exception):
    message: str


TRACEBACK_FORMAT = """## ❌ {error_title}

```
{traceback}
```

<b>{message}</b>"""


@contextmanager
def user_facing_error(message: str | Callable[[Exception], str]):
    try:
        yield
    except Exception as e:
        if callable(message):
            raise UserFacingError(message(e)) from e
        else:
            raise UserFacingError(message) from e


@dataclass
class IssueErrorHandler:
    issue: Issue
    error_title: str = "Error"
    comment: IssueComment | PullRequestComment | None = None
    tracking_id: str | None = None
    do_delete: bool = False

    def format_error(self, e: Exception, tb: str):
        mailto_link = f'<a href="mailto:support@sweep.dev?subject=Sweep%20Error:%20tracking%20id%20{self.tracking_id}&body=Sending this email to let you know that Sweep encountered an error. Thank you.">via this link</a>'
        error_message = TRACEBACK_FORMAT.format(error_title=self.error_title, traceback=tb, message=str(e))
        support_message = f"\n\nSweep has encountered a runtime error unrelated to your request. Please let us know {mailto_link} or at support@sweep.dev directly."
        result = error_message + support_message
        return sanitize_string_for_github(result)

    def display_error(self, e: Exception):
        tb = traceback.format_exc()
        if self.comment:
            self.comment.edit(
                body=self.format_error(e, tb)
                + MESSAGE_MARKER
                + DOCUMENTATION_STRING
                + format_tracking_id(self.tracking_id)
            )
        else:
            create_or_edit_sweeps_github_comment(
                issue_or_pr=self.issue,
                sweeps_message=self.format_error(e, tb),
                sweeps_github_comment=None,
                tracking_id=self.tracking_id,
            )

    def delete_comment(self):
        if self.comment:
            self.comment.delete()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            if exc_type not in (KeyboardInterrupt, SystemExit, BdbQuit):  # noqa: E721
                self.display_error(exc_value)
            return False
        elif self.do_delete:
            self.delete_comment()
        return True



from __future__ import annotations

import json
import multiprocessing
import os
import time
from typing import Generator

import backoff
import requests
from boto3 import client as boto3_client
from botocore.exceptions import ClientError
from diskcache import Cache
from loguru import logger
from numpy import array, concatenate, linalg, ndarray, where
from openai import BadRequestError, RateLimitError
from tqdm import tqdm

from sweepai.config.server import (
    BATCH_SIZE,
    CACHE_DIRECTORY,
    VOYAGE_API_AWS_ENDPOINT_NAME,
    VOYAGE_API_KEY,
    VOYAGE_API_USE_AWS,
)
from sweepai.core.llm.openai_proxy import get_embeddings_client
from sweepai.core.llm.tiktoken_utils import Tiktoken
from sweepai.utils.streamable_functions import Streamable, streamable
from sweepai.utils.timer import Timer
from sweepai.web.webhook_secret import hash_sha256

# Now uses Voyage AI if available, with asymmetric embedding
# CACHE_VERSION = "v2.0.04" + "-voyage" if VOYAGE_API_KEY else ""
suffix = "-voyage-aws" if VOYAGE_API_USE_AWS else "-voyage" if VOYAGE_API_KEY else ""
CACHE_VERSION = "v2.1.1" + suffix
vector_cache = Cache(f"{CACHE_DIRECTORY}/vector_cache")  # we instantiate a singleton, diskcache will handle concurrency


def cosine_similarity(a, B):
    from scipy.spatial.distance import cdist

    return 1 - cdist(a, B, metric="cosine")


def chunk(texts: list[str], batch_size: int) -> Generator[list[str], None, None]:
    logger.info(f"Truncating {len(texts)} texts")
    texts = [text[:25000] if len(text) > 25000 else text for text in texts]
    # remove empty string
    texts = [text if text else " " for text in texts]
    logger.info(f"Finished truncating {len(texts)} texts")
    for i in range(0, len(texts), batch_size):
        yield texts[i : i + batch_size] if i + batch_size < len(texts) else texts[i:]


# @file_cache(ignore_params=["texts"])
@streamable
def multi_get_query_texts_similarity(queries: list[str], documents: list[str]) -> list[float]:
    if not documents:
        logger.error(f"No documents to embed")
        yield "No documents to embed", []
        return []
    for message, embeddings in embed_text_array.stream(documents):
        yield message, []
    embeddings = concatenate(embeddings)
    with Timer() as timer:
        query_embedding = array(openai_call_embedding(queries, input_type="query"))
    logger.info(f"Embedding query took {timer.time_elapsed:.2f} seconds")
    with Timer() as timer:
        similarity = cosine_similarity(query_embedding, embeddings)
    logger.info(f"Similarity took {timer.time_elapsed:.2f} seconds")
    similarity = similarity.tolist()
    yield "Similarity computed.", similarity
    return similarity


def normalize_l2(x):
    x = array(x)
    if x.ndim == 1:
        norm = linalg.norm(x)
        if norm == 0:
            return x
        return x / norm
    else:
        norm = linalg.norm(x, 2, axis=1, keepdims=True)
        return where(norm == 0, x, x / norm)


def batch_by_token_count_for_voyage(
    texts: list[str],
    max_tokens: int = 120_000,
    max_length: int = 128,
) -> list[list[str]]:
    """
    This function splits the texts into batches based on the token count.
    Max token count for Voyage is 120k and max batch length count is 128.
    """
    from voyageai import Client

    client = Client()
    batches = []
    batch = []
    token_count = 0
    for text in texts:
        text_token_count = client.count_tokens([text])
        if token_count + text_token_count > max_tokens * 0.95 or len(batch) >= max_length:
            batches.append(batch)
            batch = [text]  # Start the new batch with the current text
            token_count = text_token_count  # Reset token count for the new batch
        else:
            batch.append(text)
            token_count += text_token_count
    if batch:
        batches.append(batch)
    del client
    return batches


# lru_cache(maxsize=20)
# @redis_cache()
@streamable
def embed_text_array(texts: list[str]) -> Streamable[list[ndarray]]:
    embeddings = []
    texts = [text if text else " " for text in texts]
    batches = [texts[i : i + BATCH_SIZE] for i in range(0, len(texts), BATCH_SIZE)]
    workers = min(max(1, multiprocessing.cpu_count() // 4), 1)
    last_time = time.time()
    with Timer() as timer:
        if workers > 1 and len(batches) > 1:
            with multiprocessing.Pool(processes=workers) as pool:
                embeddings = []
                for batch in tqdm(batches, total=len(batches), desc="openai embedding"):
                    embedding = pool.apply_async(openai_with_expo_backoff, (batch,))
                    embeddings.append(embedding.get())
                    if time.time() - last_time > 1:
                        yield f"{len(embeddings)}/{len(batches)} batches embedded for vector search", embeddings
                        last_time = time.time()
        else:
            # embeddings = [openai_with_expo_backoff(batch) for batch in tqdm(batches, desc="openai embedding")]
            embeddings = []
            for batch in tqdm(batches, desc="openai embedding"):
                embedding = openai_with_expo_backoff(batch)
                embeddings.append(embedding)
                if time.time() - last_time > 1:
                    yield f"{len(embeddings)}/{len(batches)} batches embedded for vector search", embeddings
                    last_time = time.time()
    logger.info(f"Embedding docs took {timer.time_elapsed:.2f} seconds")
    yield "Embedding complete.", embeddings
    return embeddings


# @redis_cache()
def openai_call_embedding_router(batch: list[str], input_type: str = "document"):  # input_type can be query or document
    from voyageai import Client

    VOYAGE_API_KEY = os.environ.get("VOYAGE_API_KEY", None)
    VOYAGE_API_AWS_ACCESS_KEY = os.environ.get("VOYAGE_API_AWS_ACCESS_KEY", None)
    VOYAGE_API_AWS_SECRET_KEY = os.environ.get("VOYAGE_API_AWS_SECRET_KEY", None)
    VOYAGE_API_AWS_REGION = os.environ.get("VOYAGE_API_AWS_REGION", None)
    VOYAGE_API_USE_AWS = VOYAGE_API_AWS_ACCESS_KEY and VOYAGE_API_AWS_SECRET_KEY and VOYAGE_API_AWS_REGION
    if len(batch) == 0:
        return array([])
    if VOYAGE_API_USE_AWS:
        sm_runtime = boto3_client(
            "sagemaker-runtime",
            aws_access_key_id=VOYAGE_API_AWS_ACCESS_KEY,
            aws_secret_access_key=VOYAGE_API_AWS_SECRET_KEY,
            region_name=VOYAGE_API_AWS_REGION,
        )
        input_json = json.dumps({"input": batch, "input_type": input_type, "truncation": "true"})
        response = sm_runtime.invoke_endpoint(
            EndpointName=VOYAGE_API_AWS_ENDPOINT_NAME,
            ContentType="application/json",
            Accept="application/json",
            Body=input_json,
        )
        body = response["Body"]
        obj = json.load(body)
        data = obj["data"]
        return array([vector["embedding"] for vector in data])
    elif VOYAGE_API_KEY:
        client = Client(api_key=VOYAGE_API_KEY)
        result = client.embed(batch, model="voyage-code-2", input_type=input_type, truncation=True)
        cut_dim = array([data for data in result.embeddings])
        normalized_dim = normalize_l2(cut_dim)
        del client
        return normalized_dim
    else:
        client = get_embeddings_client()
        model = "text-embedding-3-small"
        if client.__class__.__name__ == "AzureOpenAI":
            model = "text-embedding-3-small-1"
        response = client.embeddings.create(input=batch, model=model, encoding_format="float")
        cut_dim = array([data.embedding for data in response.data])[:, :512]
        normalized_dim = normalize_l2(cut_dim)
        # save results to redis
        return normalized_dim


def openai_call_embedding(batch: list[str], input_type: str = "document"):
    # Backoff on batch size by splitting the batch in half.
    try:
        return openai_call_embedding_router(batch, input_type)
    except ClientError as e:  # full error is botocore.errorfactory.ModelError: but I can't find it
        tiktoken_client = Tiktoken()
        if len(batch) > 1 and "Please lower the number of tokens in the batch." in str(e):
            logger.error(
                f"Token count exceeded for batch: {max([tiktoken_client.count(text) for text in batch])} retrying by splitting batch in half."
            )
            mid = len(batch) // 2
            left = openai_call_embedding(batch[:mid], input_type)
            right = openai_call_embedding(batch[mid:], input_type)
            return concatenate((left, right))
        else:
            raise e
    except BadRequestError as e:
        # In the future we can better handle this by averaging the embeddings of the split batch
        tiktoken_client = Tiktoken()
        if "maximum context length" in str(e):
            logger.warning(
                f"Token count exceeded for batch: {max([tiktoken_client.count(text) for text in batch])} truncating down to 8192 tokens."
            )
            batch = [tiktoken_client.truncate_string(text) for text in batch]
            return openai_call_embedding(batch, input_type)
        raise e


# TODO: unit test this
@backoff.on_exception(
    backoff.expo,
    (requests.exceptions.Timeout, RateLimitError),
    factor=5,
    max_tries=10,
    on_backoff=lambda details: logger.warning(
        f"Backing off {details.wait} seconds after {details.tries} tries. Total time elapsed: {details.elapsed}"
    ),
)
def openai_with_expo_backoff(batch: tuple[str]):
    # check cache first
    embeddings: list[ndarray | None] = [None] * len(batch)
    cache_keys = [hash_sha256(text) + CACHE_VERSION for text in batch]

    try:
        for i, cache_key in enumerate(cache_keys):
            cache_value = vector_cache.get(cache_key)
            # simulated_cache_miss = random.random() < 0.03
            simulated_cache_miss = False
            if cache_value is not None and not simulated_cache_miss:
                embeddings[i] = cache_value
    except Exception as e:
        logger.warning(f"Error reading embeddings from cache: {e}")

    # not stored in cache, call openai
    batch = [
        text for i, text in enumerate(batch) if embeddings[i] is None
    ]  # remove all the cached values from the batch
    if len(batch) == 0:
        embeddings = array(embeddings)
        return embeddings  # all embeddings are in cache
    try:
        # make sure all token counts are within model params (max: 8192)
        new_embeddings = openai_call_embedding(batch)
    except requests.exceptions.Timeout as e:
        logger.exception(f"Timeout error occured while embedding: {e}")
    except Exception as e:
        logger.warning(e)
        tiktoken_client = Tiktoken()
        if any(tiktoken_client.count(text) > 8192 for text in batch):
            logger.warning(
                f"Token count exceeded for batch: {max([tiktoken_client.count(text) for text in batch])} truncating down to 8192 tokens."
            )
            batch = [tiktoken_client.truncate_string(text) for text in batch]
            new_embeddings = openai_call_embedding(batch)
        else:
            raise e
    # get all indices where embeddings are None
    indices = [i for i, emb in enumerate(embeddings) if emb is None]
    # store the new embeddings in the correct position
    assert len(indices) == len(new_embeddings)
    for i, index in enumerate(indices):
        embeddings[index] = new_embeddings[i]
    # store in cache
    try:
        for cache_key, embedding in zip(cache_keys, embeddings):
            vector_cache.set(cache_key, embedding)
        embeddings = array(embeddings)
    except Exception as e:
        logger.warning(f"Error storing embeddings in cache: {e}")
    return embeddings


if __name__ == "__main__":
    texts = ["sasxtt " * 10000 for i in range(10)] + ["abb " * 1 for i in range(10)]
    embeddings = embed_text_array(texts)



from __future__ import annotations

import ast
import os
import re
import subprocess
import tempfile
import traceback
import uuid
import warnings
from dataclasses import dataclass
from io import StringIO
from tempfile import TemporaryDirectory
from typing import Optional

import tree_sitter_go
from loguru import logger
from tree_sitter import Language, Node, Parser
from tree_sitter_languages import get_language as tree_sitter_get_language

from sweepai.core.entities import Snippet
from sweepai.modify.validate.fuzzy_diff import patience_fuzzy_additions
from sweepai.utils.cache import file_cache

warnings.simplefilter("ignore", category=FutureWarning)

AVG_CHAR_IN_LINE = 60

EXT_TO_LANGUAGE = {
    "py": "python",
    "rs": "rust",
    "go": "go",
    "java": "java",
    "cpp": "cpp",
    "cc": "cpp",
    "cxx": "cpp",
    "c": "cpp",
    "h": "cpp",
    "js": "typescript",  # use typescript parser for js
    "jsx": "typescript",
    "ts": "typescript",
    "tsx": "typescript",
}


def get_language(language: str):
    language = EXT_TO_LANGUAGE.get(language, language)
    language_dict = {
        "python": tree_sitter_get_language("python"),
        "javascript": tree_sitter_get_language("javascript"),
        "typescript": tree_sitter_get_language("tsx"),
        "go": Language(tree_sitter_go.language(), "go"),  # needed for latest go grammar
        "ruby": tree_sitter_get_language("ruby"),
        "java": tree_sitter_get_language("java"),  # java implemented differently
    }
    return language_dict.get(language, None)


def get_parser(language: str):
    lang = get_language(language)
    if not lang:
        lang = tree_sitter_get_language(language)
    parser = Parser()
    parser.set_language(lang)
    return parser


def non_whitespace_len(s: str) -> int:  # new len function
    return len(re.sub("\s", "", s))


def get_line_number(index: int, source_code: str) -> int:
    total_chars = 0
    line_number = 0
    for line_number, line in enumerate(source_code.splitlines(keepends=True), start=1):
        total_chars += len(line)
        if total_chars > index:
            return line_number - 1
    return line_number


@dataclass
class Span:
    # Represents a slice of a string
    start: int = 0
    end: int = 0

    def __post_init__(self):
        # If end is None, set it to start
        if self.end is None:
            self.end = self.start

    def extract(self, s: str) -> str:
        # Grab the corresponding substring of string s by bytes
        return s[self.start : self.end]

    def extract_lines(self, s: str) -> str:
        # Grab the corresponding substring of string s by lines
        return "\n".join(s.splitlines()[self.start : self.end + 1])

    def __add__(self, other: Span | int) -> Span:
        # e.g. Span(1, 2) + Span(2, 4) = Span(1, 4) (concatenation)
        # There are no safety checks: Span(a, b) + Span(c, d) = Span(a, d)
        # and there are no requirements for b = c.
        if isinstance(other, int):
            return Span(self.start + other, self.end + other)
        elif isinstance(other, Span):
            return Span(self.start, other.end)
        else:
            raise NotImplementedError()

    def __len__(self) -> int:
        # i.e. Span(a, b) = b - a
        return self.end - self.start


def chunk_tree(
    tree,
    source_code: bytes,
    MAX_CHARS=AVG_CHAR_IN_LINE * 200,  # 200 lines of code
    coalesce=AVG_CHAR_IN_LINE * 50,  # 50 lines of code
) -> list[Span]:
    # 1. Recursively form chunks based on the last post (https://docs.sweep.dev/blogs/chunking-2m-files)
    def chunk_node(node: Node) -> list[Span]:
        chunks: list[Span] = []
        current_chunk: Span = Span(node.start_byte, node.start_byte)
        node_children = node.children
        for child in node_children:
            if child.end_byte - child.start_byte > MAX_CHARS:
                chunks.append(current_chunk)
                current_chunk = Span(child.end_byte, child.end_byte)
                chunks.extend(chunk_node(child))
            elif child.end_byte - child.start_byte + len(current_chunk) > MAX_CHARS:
                chunks.append(current_chunk)
                current_chunk = Span(child.start_byte, child.end_byte)
            else:
                current_chunk += Span(child.start_byte, child.end_byte)
        chunks.append(current_chunk)
        return chunks

    chunks = chunk_node(tree.root_node)

    # 2. Filling in the gaps
    if len(chunks) == 0:
        return []
    if len(chunks) < 2:
        end = get_line_number(chunks[0].end, source_code)
        return [Span(0, end)]
    for i in range(len(chunks) - 1):
        chunks[i].end = chunks[i + 1].start
    chunks[-1].end = tree.root_node.end_byte

    # 3. Combining small chunks with bigger ones
    new_chunks = []
    current_chunk = Span(0, 0)
    for chunk in chunks:
        current_chunk += chunk
        # if the current chunk starts with a closing parenthesis, bracket, or brace, we coalesce it with the previous chunk
        stripped_contents = current_chunk.extract(source_code.decode("utf-8")).strip()
        first_char = stripped_contents[0] if stripped_contents else ""
        if first_char in [")", "}", "]"] and new_chunks:
            new_chunks[-1] += chunk
            current_chunk = Span(chunk.end, chunk.end)
        # if the current chunk is too large, create a new chunk, otherwise, combine the chunks
        elif non_whitespace_len(
            current_chunk.extract(source_code.decode("utf-8"))
        ) > coalesce and "\n" in current_chunk.extract(source_code.decode("utf-8")):
            new_chunks.append(current_chunk)
            current_chunk = Span(chunk.end, chunk.end)
    if len(current_chunk) > 0:
        new_chunks.append(current_chunk)

    # 4. Changing line numbers
    first_chunk = new_chunks[0]
    line_chunks = [Span(0, get_line_number(first_chunk.end, source_code))]
    for chunk in new_chunks[1:]:
        start_line = get_line_number(chunk.start, source_code) + 1
        end_line = get_line_number(chunk.end, source_code)
        line_chunks.append(Span(start_line, max(start_line, end_line)))

    # 5. Eliminating empty chunks
    line_chunks = [chunk for chunk in line_chunks if len(chunk) > 0]

    # 6. Coalescing last chunk if it's too small
    if len(line_chunks) > 1 and len(line_chunks[-1]) < coalesce:
        line_chunks[-2] += line_chunks[-1]
        line_chunks.pop()

    return line_chunks


extension_to_language = {
    "js": "tsx",
    "jsx": "tsx",
    "ts": "tsx",
    "tsx": "tsx",
    "mjs": "tsx",
    "py": "python",
    "rs": "rust",
    "go": "go",
    "java": "java",
    "cpp": "cpp",
    "cc": "cpp",
    "cxx": "cpp",
    "c": "cpp",
    "h": "cpp",
    "hpp": "cpp",
    "cs": "cpp",
    "rb": "ruby",
    # "md": "markdown",
    # "rst": "markdown",
    # "txt": "markdown",
    # "erb": "embedded-template",
    # "ejs": "embedded-template",
    # "html": "embedded-template",
    "erb": "html",
    "ejs": "html",
    "html": "html",
    "vue": "html",
    "php": "php",
    "elm": "elm",
}


def naive_chunker(code: str, line_count: int = 30, overlap: int = 15):
    if overlap >= line_count:
        raise ValueError("Overlap should be smaller than line_count.")
    lines = code.split("\n")
    total_lines = len(lines)
    chunks = []

    start = 0
    while start < total_lines:
        end = min(start + line_count, total_lines)
        chunk = "\n".join(lines[start:end])
        chunks.append(chunk)
        start += line_count - overlap

    return chunks


def get_new_lint_errors_for_pylint(new_errors: str, old_errors: str) -> str:
    # Example of error: "main.py:1585:12: W0612: Unused variable 'command' (unused-variable)"
    additional_errors = patience_fuzzy_additions(old_errors, new_errors).splitlines()
    old_error_types = []
    for line in old_errors.splitlines():
        if line.count(" ") > 2:
            _file_delimiter, error_type, *_ = line.split(" ")
            old_error_types.append(error_type)
    results = []
    try:
        for line in additional_errors:
            if line.count(" ") >= 2:  # sometimes the error doesn't have enough spaces, which raises an error
                _file_delimiter, error_type, *_ = line.split(" ")
            else:
                _file_delimiter, error_type = line.split(" ")
            if (
                error_type.startswith("E") or old_error_types.count(error_type) < 2
            ):  # if there are more than 1 of the same error, we consider it new
                results.append(line)
    except Exception as e:
        logger.info(f"Error in get_new_lint_errors_for_pylint: {e}")
    return "\n".join(results)


def get_new_lint_errors_for_eslint(new_errors: str, old_errors: str) -> str:
    # Example of error: "main.py:1585:12: W0612: Unused variable 'command' (unused-variable)"
    additional_errors = patience_fuzzy_additions(old_errors, new_errors).splitlines()
    old_error_types = []
    for line in old_errors.splitlines():
        if line.count(" ") > 2:
            *_, error_type = line.split(" ")
            old_error_types.append(error_type)
    results = []
    for line in additional_errors:
        *_, error_type = line.split(" ")
        if (
            not line.startswith("✖") and old_error_types.count(error_type) < 2
        ):  # if there are more than 1 of the same error, we consider it new
            results.append(line)
    return "\n".join(results)


@dataclass
class CheckResults:
    # Experimental feature, we'll see how this does.
    # TODO: smart parsing
    parse_error_message: str = ""
    pylint: str = ""
    eslint: str = ""

    def is_worse_than(self, other: CheckResults) -> bool:
        if self.parse_error_message:
            return True
        if other.parse_error_message:
            return False
        return len(self.pylint.splitlines()) > len(other.pylint.splitlines()) or len(self.eslint.splitlines()) > len(
            other.eslint.splitlines()
        )

    def is_worse_than_message(self, other: CheckResults) -> str:
        if other.parse_error_message:
            # Previously failing
            return ""
        if self.parse_error_message:
            return self.parse_error_message
        if len(self.pylint.splitlines()) > len(other.pylint.splitlines()):
            # return f"The code has the following pylint errors:\n\n{self.pylint}"
            new_pylint_errors = get_new_lint_errors_for_pylint(self.pylint, other.pylint)
            if not other.pylint:
                return f"The code has the following pylint errors:\n\n{self.pylint}"
            elif not new_pylint_errors:
                # All the errors are invalid
                return ""
            return f"The following new pylint errors have appeared:\n\n{new_pylint_errors}"
        if len(self.eslint.splitlines()) > len(other.eslint.splitlines()):
            new_eslint_errors = get_new_lint_errors_for_eslint(self.eslint, other.eslint)
            if not other.eslint:
                return f"The code has the following eslint errors:\n\n{self.eslint}"
            elif not new_eslint_errors:
                # All the errors are invalid
                return ""
            return f"The following new eslint errors have appeared:\n\n{new_eslint_errors}"
        return ""


def strip_ansi_codes(text: str) -> str:
    # ANSI escape sequences (color codes) are often starting with ESC ([) followed by some numbers and ends with "m".
    ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")
    return ansi_escape.sub("", text)


def check_valid_typescript(file_path: str, code: str) -> tuple[bool, str]:
    is_valid = True
    message = ""
    version_check = ["tsc", "--version"]
    result = subprocess.run(
        " ".join(version_check),
        capture_output=True,
        text=True,
        shell=True,
    )
    # only run if tsc is available
    if result.returncode == 0:
        # Create a temporary file to hold the TypeScript code
        with tempfile.NamedTemporaryFile(suffix=".ts", delete=False) as temp_file:
            temp_file_path = temp_file.name
            temp_file.write(code.encode("utf-8"))

        # Run `tsc` on the temporary file
        try:
            commands = ["tsc", "--pretty", "--noEmit", temp_file_path]
            result = subprocess.run(" ".join(commands), shell=True, text=True, capture_output=True)

            if result.returncode != 0:
                message = strip_ansi_codes(result.stdout)
                index = message.index(temp_file_path)
                full_temp_file_path = message[: index + len(temp_file_path)]
                message = message.replace(full_temp_file_path, file_path)

                # import error is TS2307 and should come up after the syntax check
                import_error = "error TS2307"
                if import_error in message:
                    num_of_errors = message.count(import_error)
                    # see if this matches the total amount of errors:
                    total_error_message = f"Found {num_of_errors} error"
                    if total_error_message in message:
                        # if we only have import errors, we consider it a successful check
                        return True, ""
                    else:
                        # there are more errors than just import errors
                        # now attempt to parse the message so that we remove import errors
                        message_lines = message.split("\n")
                        while num_of_errors > 0:
                            for line in message_lines:
                                if import_error in line:
                                    message_lines = message_lines[5:]
                                    num_of_errors -= 1
                        message = "\n".join(message_lines)
                return False, message
        finally:
            # Clean up: remove the temporary file
            os.remove(temp_file_path)
    return is_valid, message


def check_syntax(file_path: str, code: str) -> tuple[bool, str]:
    ext = file_path.split(".")[-1]
    if ext in extension_to_language:
        language = extension_to_language[ext]
    else:
        return True, "Unsupported file extension, skipping syntax check."
    parser = get_parser(language)
    tree = parser.parse(code.encode("utf-8"))

    if language == "python":
        # First check for syntax errors
        try:
            ast.parse(code)
        except SyntaxError as e:
            error_message = f"Python syntax error: {e.msg} at line {e.lineno}"
            return False, error_message

    # we can't do this right now unfortunately as we don't have a way to mimic the production env for the code
    # if ext in ["ts"]:
    #     return check_valid_typescript(file_path, code)

    def find_deepest_error(node: Node) -> Optional[Node]:
        deepest_error = None
        if node.has_error:
            deepest_error = node
        for child in node.children:
            child_error = find_deepest_error(child)
            if child_error:
                deepest_error = child_error
        return deepest_error

    error_location = find_deepest_error(tree.root_node)
    if error_location:
        start_line, start_col = error_location.start_point
        end_line, end_col = error_location.end_point
        code_lines = code.split("\n")
        surrounding_lines = 3
        error_code_lines = code_lines[max(0, start_line - surrounding_lines) : start_line]
        if start_line == end_line:
            error_code_lines += [code_lines[start_line]]
            error_code_lines += [" " * start_col + "^" * max(end_col - start_col, 1)]
        else:
            error_code_lines += ["=== ERROR START ==="]
            error_code_lines += code_lines[start_line : end_line + 1]
            error_code_lines += ["=== ERROR END ==="]
        error_code_lines += code_lines[end_line + 1 : min(len(code_lines) - 1, end_line + surrounding_lines)]
        error_span = "\n".join(error_code_lines)
        if start_line == end_line:
            error_message = f"Invalid syntax found at line {start_line}, displayed below:\n{error_span}"
        else:
            error_message = f"Invalid syntax found from {start_line}-{end_line}, displayed below:\n{error_span}"
        return (False, error_message)
    return True, ""


# Need to add "no-unused-vars": "error"
# Need to add "import/first": "error"

DEFAULT_ESLINTRC = """{
    "parser": "@typescript-eslint/parser",
    "parserOptions": {
      "ecmaVersion": 2020,
      "sourceType": "module",
      "ecmaFeatures": {
        "jsx": true
      }
    },
    "settings": {
      "react": {
        "version": "detect"
      }
    },
    "extends": [
      "eslint:recommended",
      "plugin:react/recommended",
      "plugin:@typescript-eslint/recommended"
    ],
    "env": {
      "browser": true,
      "es2021": true,
      "node": true
    },
    "plugins": [
      "react",
      "@typescript-eslint"
    ],
    "rules": {
        "no-undef": "error",
        "no-const-assign": "error",
        "no-redeclare": "error",
        "no-unused-vars": "error",
        "no-use-before-define": ["error", { "functions": true, "classes": true, "variables": true }],
        "import/first": "error"
    }
}
"""

pylint_args_non_last_fcr = [
    "--disable=C",
    "--enable=C0413",  # Enable only the check for imports not at the top
    "--disable=W0611",  # Don't check unused import
    "--disable=R",
    "--disable=import-error",
    "--disable=no-member",
]

# add a comment to all lines which are changed
pylint_args_last_fcr = [
    "--disable=C",
    "--enable=C0413",
    "--enable=W0611",  # Check unused import
    "--disable=R",
    "--disable=import-error",
    "--disable=no-member",
]


@file_cache()
def get_pylint_check_results(file_path: str, code: str, last_fcr_for_file=False) -> CheckResults:
    from pylint.lint import Run
    from pylint.reporters.text import TextReporter

    logger.debug(f"Running pylint on {file_path}...")
    file_hash = uuid.uuid4().hex
    new_file = os.path.join("/tmp", file_hash + "_" + os.path.basename(file_path))
    stem = os.path.splitext(os.path.basename(file_path))[0]
    with open(new_file, "w") as f:
        f.write(code)
    pylint_output = StringIO()
    reporter = TextReporter(pylint_output)
    # this allows us to have a more rigorous check for the last file change request
    pylint_args = [new_file] + (pylint_args_last_fcr if last_fcr_for_file else pylint_args_non_last_fcr)
    Run(
        pylint_args,
        reporter=reporter,
        exit=False,
    )
    error_message = pylint_output.getvalue().strip()
    try:
        os.remove(new_file)
    except FileNotFoundError:
        pass
    succeeded = error_message.startswith("------------------------------------")
    if error_message:
        error_message = error_message.replace(new_file, file_path).replace(f"{file_hash}_" + stem, stem)
        error_message = error_message.split("-----------------------------------", 1)[0].strip()
        error_message = f"> pylint {file_path}\n\n" + error_message
    logger.debug("Done running pylint.")
    return CheckResults(pylint=error_message if not succeeded else "")


# TODO: unit test this
def get_check_results(file_path: str, code: str, last_fcr_for_file=False) -> CheckResults:
    is_valid, error_message = check_syntax(file_path, code)
    if not is_valid:
        return CheckResults(parse_error_message=error_message)
    ext = file_path.rsplit(".")[-1]  # noqa
    if ext == "py":
        try:
            return get_pylint_check_results(file_path, code, last_fcr_for_file=last_fcr_for_file)
        except Exception as e:
            logger.exception(e)
    elif ext in ["js", "jsx", "ts", "tsx"]:
        # see if eslint is installed
        npx_commands = ["npx", "eslint", "--version"]
        try:
            result = subprocess.run(
                " ".join(npx_commands),
                timeout=5,
                capture_output=True,
                text=True,
                shell=True,
            )
        except subprocess.TimeoutExpired:
            raise Exception(
                "ESLint timed out after 5s. You need eslint to edit js/ts files. Run `npm i -g eslint @typescript-eslint/parser @typescript-eslint/eslint-plugin eslint-plugin-import eslint-plugin-react`."
            )
        # Check eslint < v9 and all the plugins exist
        if result.returncode == 0:
            with TemporaryDirectory(dir=os.getcwd()) as temp_dir:
                file_name = file_path.split(os.path.sep)[-1]
                new_file = os.path.join(temp_dir, f"{file_name}")
                config_file = os.path.join(temp_dir, ".eslintrc")
                with open(config_file, "w") as f:
                    f.write(DEFAULT_ESLINTRC)
                with open(new_file, "w") as f:
                    f.write(code)
                try:
                    eslint_commands = [
                        "npx",
                        "eslint",
                        new_file,
                        "--config",
                        config_file,
                        "--no-ignore",
                    ]
                    result = subprocess.run(
                        " ".join(eslint_commands),
                        capture_output=True,
                        text=True,
                        shell=True,
                        timeout=30,
                    )
                    error_message = (result.stdout + "\n\n" + result.stderr).strip().replace(new_file, file_path)
                    return CheckResults(eslint=error_message)
                except subprocess.TimeoutExpired:
                    logger.warning(f"ESLint timed out after 30s for {file_path}")
                    pass
    return CheckResults()


PRETTIERRC_FILES = [
    ".prettierrc",
    ".prettierrc.json",
    ".prettierrc.yaml",
    ".prettierrc.yml",
    ".prettierrc.js",
    "prettier.config.js",
    "package.json",
]

CLANG_FORMAT_CONFIG_FILES = [
    ".clang-format",
]


def format_file(file_path: str, code: str, cwd: str | None = None) -> str:
    """
    Format the given code based on the file extension.
    Currently supports JavaScript, TypeScript, JSX, Go, Python, Java, and C++.
    """
    _, ext = os.path.splitext(file_path)
    ext = ext.removeprefix(".")

    if ext in ("js", "jsx", "ts", "tsx"):
        return format_with_prettier(file_path, code, cwd)
    elif ext == "go":
        return format_with_gofmt(code)
    elif ext == "py":
        return format_with_ruff(code, cwd)
    elif ext == "java":
        return format_with_google_java_formatter(code)
    elif ext in ("cpp", "cc", "cxx", "c", "h", "hpp"):
        return format_with_clang_format(code, cwd)
    return code


def format_with_prettier(file_path: str, code: str, cwd: str | None) -> str:
    prettier_config = find_prettier_config(cwd)
    if not prettier_config:
        return code

    with TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        config_file = os.path.join(temp_dir, prettier_config["file"])
        with open(config_file, "w") as f:
            f.write(prettier_config["contents"])

        npx_commands = ["prettier", "--stdin-filepath", file_path]
        return run_formatter(npx_commands, code, temp_dir)


def format_with_gofmt(code: str) -> str:
    gofmt_command = ["gofmt"]
    return run_formatter(gofmt_command, code)


def format_with_ruff(code: str, cwd: str | None = None) -> str:
    extension = []
    if cwd and os.path.exists(os.path.join(cwd, "ruff.toml")):
        extension = ["--config", os.path.join(cwd, "ruff.toml")]
    else:
        extension = ["--no-cache"]

    ruff_check_command = ["ruff", "check", "--fix", "-", *extension]
    checked_results = run_formatter(ruff_check_command, code, cwd=cwd)
    ruff_format_command = ["ruff", "format", "-", *extension]
    return run_formatter(ruff_format_command, checked_results, cwd=cwd)


def format_with_google_java_formatter(code: str) -> str:
    with tempfile.NamedTemporaryFile(suffix=".java", mode="w+", delete=False) as temp_file:
        temp_file.write(code)
        temp_file_path = temp_file.name

    try:
        formatter_command = [
            "google-java-format",
            "--aosp",
            "--replace",
            temp_file_path,
        ]
        subprocess.run(formatter_command, check=True, capture_output=True, text=True)

        with open(temp_file_path, "r") as file:
            formatted_code = file.read()

        return formatted_code
    except subprocess.CalledProcessError as e:
        logger.warning(f"Google Java Formatter error: {e.stderr}")
    except Exception as e:
        logger.warning(f"Unexpected error during Java formatting: {e}")
    finally:
        os.remove(temp_file_path)

    return code


def find_clang_format_config(cwd: str | None) -> dict | None:
    if not cwd:
        return None

    for config_file in CLANG_FORMAT_CONFIG_FILES:
        full_path = os.path.join(cwd, config_file)
        if os.path.exists(full_path):
            with open(full_path, "r") as f:
                return {"file": config_file, "contents": f.read()}

    return None


def format_with_clang_format(code: str, cwd: str) -> str:
    clang_format_config = find_clang_format_config(cwd)
    if not clang_format_config:
        return code

    with tempfile.NamedTemporaryFile(suffix=".cpp", mode="w+", delete=False) as temp_file:
        temp_file.write(code)
        temp_file_path = temp_file.name

    try:
        formatter_command = ["clang-format", "-i", temp_file_path]
        subprocess.run(formatter_command, check=True, capture_output=True, text=True)

        with open(temp_file_path, "r") as file:
            formatted_code = file.read()

        return formatted_code
    except subprocess.CalledProcessError as e:
        logger.warning(f"clang-format error: {e.stderr}")
    except Exception as e:
        logger.warning(f"Unexpected error during C++ formatting: {e}")
    finally:
        os.remove(temp_file_path)

    return code


def find_prettier_config(cwd: str | None) -> dict | None:
    if not cwd:
        return None

    for config_file in PRETTIERRC_FILES:
        full_path = os.path.join(cwd, config_file)
        if os.path.exists(full_path):
            with open(full_path, "r") as f:
                return {"file": config_file, "contents": f.read()}
    return None


def run_formatter(command: list[str], code: str, cwd: str | None = None) -> str:
    try:
        result = subprocess.run(
            command,
            input=code,
            capture_output=True,
            text=True,
            check=True,
            cwd=cwd,
        )
        return result.stdout
    except subprocess.CalledProcessError as e:
        logger.warning(f"Formatting error: {e.stderr}")
    except Exception as e:
        logger.warning(f"Unexpected error during formatting: {e}")
    return code


# probably add a way to do this langauge agnostic
PLACEHOLDER_REGEXES = [
    # matches cases like // ... rest of the code
    # making this relatively strict but there's no reason we can't ban all // ... comments
    r"(//|#) \.\.\..*(rest of the)",
    r"(//|#) \.\.\..*(existing)",
    r"(//|#) \.\.\..*(other methods)",
    r"(//|#) \.\.\..*(other fields)",
    r"(//|#) \.\.\..*(other classes)",
    r"(//|#) \.\.\..*(other functions)",
    r"(//|#) \.\.\..*(implementation)",
    r"(//|#) \.\.\..*(import)",
]


def remove_placeholder_comments_from_code(code: str) -> str:
    code_without_placeholder_comments = []
    for line in code.split("\n"):
        if any(re.match(placeholder_regex, line.strip()) for placeholder_regex in PLACEHOLDER_REGEXES):
            continue
        else:
            code_without_placeholder_comments.append(line)
    return "\n".join(code_without_placeholder_comments)


# @file_cache()
def chunk_code(
    code: str,
    path: str,
    MAX_CHARS=AVG_CHAR_IN_LINE * 100,  # 100 lines of code
    coalesce=80,
) -> list[Snippet]:
    ext = path.split(".")[-1]
    if ext in extension_to_language:
        language = extension_to_language[ext]
    else:
        # Fallback to naive chunking if tree_sitter fails
        line_count = MAX_CHARS // AVG_CHAR_IN_LINE
        overlap = 0
        chunks = naive_chunker(code, line_count, overlap)
        snippets = []
        for idx, chunk in enumerate(chunks):
            end = min((idx + 1) * (line_count - overlap), len(code.split("\n")))
            new_snippet = Snippet(
                content=code,
                start=idx * (line_count - overlap),
                end=end,
                file_path=path,
            )
            snippets.append(new_snippet)
        return snippets
    try:
        parser = get_parser(language)
        tree = parser.parse(code.encode("utf-8"))
        chunks = chunk_tree(tree, code.encode("utf-8"), MAX_CHARS=MAX_CHARS, coalesce=coalesce)
        snippets = []
        for chunk in chunks:
            new_snippet = Snippet(
                content=code,
                start=chunk.start,
                end=chunk.end,
                file_path=path,
            )
            snippets.append(new_snippet)
        return snippets
    except Exception:
        logger.error(traceback.format_exc())
        return []


def get_function_name(file_name: str, source_code: str, line_number: int):
    ext = file_name.split(".")[-1]
    if ext in extension_to_language:
        language = extension_to_language[ext]
    else:
        return None
    type_mapping_per_language = {
        "python": "function_definition",
        "tsx": "function_declaration",
        "js": "function_declaration",
    }
    function_type_string = type_mapping_per_language.get(language)
    parser = get_parser(language)

    # Parse the source code
    tree = parser.parse(bytes(source_code, "utf8"))

    # Get the root node of the syntax tree
    root_node = tree.root_node

    # Find the function node that contains the given line number
    function_node = root_node.descendant_for_point_range((line_number, 0), (line_number, 1))

    max_depth = 25  # Maximum depth to search for the function node
    while function_node.type != function_type_string:
        function_node = function_node.parent
        max_depth -= 1
        if max_depth == 0 or function_node is None:
            return None

    # Extract the function name
    function_name = function_node.child_by_field_name("name").text.decode("utf8")

    return function_name


def extract_definitions(file_name, code):
    ext = file_name.split(".")[-1]
    if ext in extension_to_language:
        language = extension_to_language[ext]
    else:
        return None

    type_mapping_per_language = {
        "typescript": ["class_declaration", "method_definition"],
        "tsx": ["function_declaration", "class_declaration", "method_definition"],
        "javascript": [
            "function_declaration",
            "class_declaration",
            "method_definition",
        ],
    }
    function_type_strings = type_mapping_per_language.get(language)

    parser = get_parser(language)
    tree = parser.parse(bytes(code, "utf8"))

    def traverse_node(node):
        if node.type in function_type_strings:
            if node.type == "class_declaration":
                class_name = node.child_by_field_name("name").text.decode("utf8")
                print(f"Class: {class_name}")
            elif node.type == "function_declaration":
                function_name = node.child_by_field_name("name").text.decode("utf8")
                print(f"Function: {function_name}")
            elif node.type == "method_definition":
                method_name = node.child_by_field_name("name").text.decode("utf8")
                print(f"Method: {method_name}")

        for child in node.children:
            traverse_node(child)

    traverse_node(tree.root_node)


if __name__ == "__main__":
    # TODO: move this to unit tests
    test_code = """
import React from 'react';
import { render } from '@testing-library/react';
import CallToAction from '../components/CallToAction';
describe('CallToAction component', () => {
  it('renders the correct YouTube video link', () => {
    const { getByTitle } = render(<CallToAction />);
    const iframeElement = getByTitle('YouTube video player');
    expect(iframeElement.getAttribute('src')).toBe('https://www.youtube.com/embed/GVEkDZmWw8E?autoplay=1&mute=1&loop=1&vq=hd1080&modestbranding=1&controls=0');
    it('has a button with the correct color properties', () => {
    const { getByRole } = render(<CallToAction />);
    const buttonElement = getByRole('button', { name: /install sweep/i });
    expect(buttonElement).toHaveStyle({
      colorScheme: 'green',
      bg: 'green.400',
      _hover: { bg: 'green.600' }
    });
  });
});
"""

    test_code = """
x = "test"

import numpy
"""
    typescript_code = """import {
    Flex,
    Container,
    Heading,
    Stack,
    Text,
    Button,
  } from "@chakra-ui/react";
  import { tsParticles } from "tsparticles";
  import { loadConfettiPreset } from "tsparticles-preset-confetti";
  import { useState } from "react";
  import logo from "../assets/icon.png";
  
  import ExternalLinkWithText from "./ExternalLinkWithText";
  import { TypeAnimation } from "react-type-animation";
  import { FontAwesomeIcon } from '@fortawesome/react-fontawesome';
  import { faUsers } from '@fortawesome/free-solid-svg-icons';
  const demo = require("../assets/demo.mp4");
  
export default function CallToAction() {
    const [spin, setSpin] = useState(false);
    // const canvas = document.getElementById('canvas3d');
    // const app = new Application(canvas);
    // app.load('https://prod.spline.design/jzV1MbbHCyCmMG7u/scene.splinecode');
    return (
      <Container maxW={"5xl"}>
      </Container>
    );
}
"""
    extract_definitions("main.tsx", typescript_code)
    python_code = """\
import math
import pandas

def get_circle_area(radius: float) -> float:
    return math.pi * radius ** 2
"""
    print(get_check_results("main.py", test_code))
    formatted_code = format_file("main.tsx", typescript_code, cwd="sweep_chat")
    function_name = get_function_name("main.ts", typescript_code, 20)
    print(function_name)
    function_name = get_function_name("main.py", python_code, 3)
    print(function_name)
    # new_code = """console.log("hello world")"""
    # check_results = check_syntax("test.js", new_code)
    check_results = get_check_results("test.tsx", typescript_code)
    check_results = get_check_results("test.py", python_code)
    assert check_results.pylint == ""  # this should pass
    check_results = get_check_results("test.py", python_code, last_fcr_for_file=True)
    assert "Unused import pandas" in check_results.pylint  # this should warn about unused imports



import re
from typing import Any


def xml_pattern(
    tags: str,
    name: str | None = None,
    add_newlines: bool = True,
    **kwargs: dict[str, str],
) -> str:
    name = name or tags
    new_lines = "\n" if add_newlines else ""
    if kwargs:
        kwargs_pattern = r"\s+" + r"\s+".join(rf"{key}=\"(?P<{value}>.*?)\"" for key, value in kwargs.items())
    else:
        kwargs_pattern = ""
    return rf"<{tags}{kwargs_pattern}>{new_lines}(?P<{name}>.*?){new_lines}</{tags}>"


tool_call_parameters = {
    "make_change": [
        "justification",
        "file_name",
        "original_code",
        "new_code",
        "replace_all",
    ],
    "create_file": ["justification", "file_name", "new_code"],
    "submit_task": ["sources", "justification", "answer"],
    "search_codebase": ["query", "question", "include_docs", "include_tests"],
    "semantic_search": ["query", "question", "include_docs", "include_tests"],
    "ripgrep": ["query"],
}


# returns a dictionary of the tool call parameters, assumes correct
def parse_function_call_parameters(tool_call_contents: str, parameters: list[str]) -> dict[str, Any]:
    tool_args = {}
    for param in parameters:
        param_regex = rf"<{param}>(?P<{param}>.*?)<\/{param}>"
        match = re.search(param_regex, tool_call_contents, re.DOTALL)
        if match:
            param_contents = match.group(param)
            tool_args[param] = param_contents
    return tool_args


# parse llm response for tool calls in xml format
def parse_function_calls(response_contents: str) -> list[dict[str, str]]:
    tool_calls = []
    # first get all tool calls
    for tool_name in tool_call_parameters.keys():
        tool_call_regex = rf"<{tool_name}>(?P<function_call>.*?)<\/{tool_name}>"
        tool_call_matches = re.finditer(tool_call_regex, response_contents, re.DOTALL)
        # now we extract its parameters
        for tool_call_match in tool_call_matches:
            tool_call_contents = tool_call_match.group("function_call")
            # get parameters based off of tool name
            parameters = tool_call_parameters[tool_name]
            tool_call = {
                "tool": tool_name,
                "arguments": parse_function_call_parameters(tool_call_contents, parameters),
                "raw_text": tool_call_match.group(0),
            }
            tool_calls.append(tool_call)
    return tool_calls


if __name__ == "__main__":
    pattern = xml_pattern("additional_changes", required="additional_changes_required")
    print(pattern)
    example_template = """\
<additional_changes required="yes">
Test
</additional_changes>"""
    print(re.match(pattern, example_template))



from github.Commit import Commit
from github.GithubException import GithubException
from github.InputGitTreeElement import InputGitTreeElement
from loguru import logger

from sweepai.config.server import GITHUB_BASE_URL
from sweepai.core.github_utils import github_integration
from sweepai.o11y.log_utils import log_to_file, log_warnings_to_file
from sweepai.utils.file_utils import safe_decode
from sweepai.utils.ticket_rendering_utils import (
    PR_REVERTED_INDICATOR,
    create_or_edit_sweeps_github_comment,
)
from sweepai.web.events import CommentCreatedOrEditedRequest, IssueCommentRequest


def extract_pr_and_commit_from_comment(original_line_to_revert: str) -> tuple[int, str]:
    # Expects the entire comment
    split_original_line = original_line_to_revert.split(PR_REVERTED_INDICATOR)[-1].split(f"](https://{GITHUB_BASE_URL}")
    assert len(split_original_line) >= 2 and isinstance(
        split_original_line[1], str
    ), f"Invalid first half of comment: {original_line_to_revert}"
    pr_url = split_original_line[1].split(")")[0]
    # index from the end to be safer
    split_pr_url = pr_url.split("/")
    assert (
        len(split_pr_url) > 2 and isinstance(split_pr_url[-1], str) and isinstance(split_pr_url[-2], str)
    ), f"Invalid PR Url/second half of comment: {split_pr_url}"
    pr_number = split_pr_url[-3]
    commit_sha = split_pr_url[-1]
    assert pr_number.isdigit() and isinstance(
        commit_sha, str
    ), f"Invalid PR number or commit SHA: {pr_number}, {commit_sha}"
    return int(pr_number), commit_sha


def create_commit_name_for_revert(commit_to_revert: Commit) -> str:
    # Create the revert commit
    # truncate message accounting 50 character limit + length of other text (which is 43 from "revert " (7)
    safe_length_message = (
        commit_to_revert.commit.message
        if len(commit_to_revert.commit.message) < 43
        else commit_to_revert.commit.message[: commit_to_revert.commit.message.rfind(" ", 0, 43)]
    )
    commit_message = f"Revert {safe_length_message}"
    assert len(commit_message) <= 50, "Commit message too long"
    return commit_message


def on_revert_commit_comment(
    issue_comment_request: IssueCommentRequest | CommentCreatedOrEditedRequest,
    tracking_id: str,
):
    with log_to_file(tracking_id), log_warnings_to_file(keywords=["[YELLOW]"], tracking_id=tracking_id):
        original_comment_id = issue_comment_request.comment.id
        current_body = issue_comment_request.comment.body
        repo_full_name = issue_comment_request.repository.full_name
        org_name, repo_name = repo_full_name.split("/")
        installation = github_integration.get_repo_installation(org_name, repo_name)
        repo = installation.get_github_for_installation().get_repo(repo_full_name)
        pr_number, commit_sha = extract_pr_and_commit_from_comment(current_body)
        pull_request = repo.get_pull(pr_number)

        if isinstance(issue_comment_request, CommentCreatedOrEditedRequest):
            all_comments = pull_request.get_review_comments()
        else:
            all_comments = pull_request.get_issue_comments()

        sweeps_github_comment = None
        for comment in all_comments:
            if comment.id == original_comment_id:
                sweeps_github_comment = comment
                break
        sweeps_github_comment = create_or_edit_sweeps_github_comment(
            issue_or_pr=pull_request,
            sweeps_message=f"I'm reverting #{pr_number}...",
            sweeps_github_comment=sweeps_github_comment,
            tracking_id=tracking_id,
        )

        commit_to_revert = repo.get_commit(commit_sha)

        # check no other commits have been pushed to the branch since the PR was created
        if commit_sha != pull_request.head.sha:
            logger.info(
                f"More commits have been pushed to the branch since the PR was created. Not reverting {commit_sha}."
            )
            return

        head_branch = pull_request.head.ref

        try:
            # Get the diff of the commit to revert
            diff = commit_to_revert.files

            commit_before_commit_to_revert = commit_to_revert.parents[0]

            # Create dictionaries to store the current state of the tree, file modes, and contents
            current_tree = {}
            file_contents = {}

            for file_to_revert in commit_to_revert.files:
                current_tree[file_to_revert.filename] = file_to_revert.sha

            # Apply the inverse changes to the current_tree dictionary
            for file in diff:
                if file.status == "added":
                    # If file was added, we'll delete it
                    continue
                elif file.status == "removed":
                    # If file was removed, we'll add it back
                    content = (
                        safe_decode(
                            repo=repo,
                            path=file.filename,
                            ref=commit_before_commit_to_revert.sha,
                        )
                        or ""
                    )
                    if content is not None:
                        current_tree[file.filename] = file.filename
                        file_contents[file.filename] = content
                elif file.status == "modified":
                    # If file was modified, we'll revert to the previous version
                    content = (
                        safe_decode(
                            repo=repo,
                            path=file.filename,
                            ref=commit_before_commit_to_revert.sha,
                        )
                        or ""
                    )
                    if content is not None:
                        current_tree[file.filename] = file.filename
                        file_contents[file.filename] = content
                elif file.status == "renamed":
                    # If file was renamed, we'll move it back
                    content = (
                        safe_decode(
                            repo=repo,
                            path=file.previous_filename,
                            ref=commit_before_commit_to_revert.sha,
                        )
                        or ""
                    )
                    if content is not None:
                        current_tree[file.previous_filename] = file.previous_filename
                        file_contents[file.previous_filename] = content

            # Create the list of InputGitTreeElements
            tree_elements = [
                InputGitTreeElement(path=path, mode="100644", type="blob", content=file_contents[path])
                for path in current_tree.keys()
            ]

            # Create a new tree with the inverse changes
            new_tree = repo.create_git_tree(tree_elements, base_tree=repo.get_git_tree(sha=pull_request.head.sha))

            # Create the revert commit
            commit_message = create_commit_name_for_revert(commit_to_revert)
            revert_commit = repo.create_git_commit(
                message=commit_message,
                tree=new_tree,
                parents=[repo.get_git_commit(sha=pull_request.head.sha)],
            )

            # Update the branch reference
            ref = repo.get_git_ref(f"heads/{head_branch}")
            ref.edit(sha=revert_commit.sha)

            logger.info(f"Successfully created revert commit on branch {head_branch}")

            successful_revert_message = f"I've reverted commit {revert_commit.sha} on branch `{head_branch}`. "
            sweeps_github_comment = create_or_edit_sweeps_github_comment(
                issue_or_pr=pull_request,
                sweeps_message=successful_revert_message,
                sweeps_github_comment=sweeps_github_comment,
                tracking_id=tracking_id,
            )
        except GithubException as e:
            logger.error(f"GitHub API error while creating revert commit: {str(e)}")
            sweeps_github_comment = create_or_edit_sweeps_github_comment(
                issue_or_pr=pull_request,
                sweeps_message=f"I failed to create a revert commit. GitHub API Error: {str(e)}",
                sweeps_github_comment=sweeps_github_comment,
                tracking_id=tracking_id,
            )
        except Exception as e:
            logger.error(f"Error while creating revert commit: {str(e)}")
            sweeps_github_comment = create_or_edit_sweeps_github_comment(
                issue_or_pr=pull_request,
                sweeps_message=f"I failed to create a revert commit. Error: {str(e)}",
                sweeps_github_comment=sweeps_github_comment,
                tracking_id=tracking_id,
            )
